#!/usr/bin/env python3

import argparse
import os
import re

import didppy as dp

import read_tsplib


def solve(n, nodes, edges, capacity, demand, k, time_limit=None):
    model = dp.Model()
    customer = model.add_object_type(number=n)
    unvisited = model.add_set_var(object_type=customer, target=[i for i in range(1, n)])
    location = model.add_element_var(object_type=customer, target=0)
    load = model.add_int_resource_var(target=0, less_is_better=True)
    vehicles = model.add_int_resource_var(target=1, less_is_better=True)
    demand = model.add_int_table([demand[i] for i in nodes])
    distance_matrix = [
        [edges[i, j] if (i, j) in edges else 0 for j in nodes] for i in nodes
    ]
    distance = model.add_int_table(distance_matrix)
    distance_via_depot = model.add_int_table(
        [
            [
                edges[i, nodes[0]] + edges[nodes[0], j]
                if (i, nodes[0]) in edges and (nodes[0], j) in edges
                else edges[i, j]
                if (i, j) in edges
                else 0
                for j in nodes
            ]
            for i in nodes
        ]
    )

    model.add_base_case([unvisited.is_empty(), location == 0])
    name_to_partial_tour = {}

    for i in range(1, n):
        name = "visit {}".format(i)
        name_to_partial_tour[name] = (nodes[i],)
        visit = dp.Transition(
            name=name,
            cost=dp.IntExpr.state_cost() + distance[location, i],
            effects=[
                (unvisited, unvisited.remove(i)),
                (location, i),
                (load, load + demand[i]),
            ],
            preconditions=[unvisited.contains(i), load + demand[i] <= capacity],
        )
        model.add_transition(visit)

    for i in range(1, n):
        name = "visit {} via depot".format(i)
        name_to_partial_tour[name] = (nodes[0], nodes[i])
        visit_via_depot = dp.Transition(
            name=name,
            cost=dp.IntExpr.state_cost() + distance_via_depot[location, i],
            effects=[
                (unvisited, unvisited.remove(i)),
                (location, i),
                (load, demand[i]),
                (vehicles, vehicles + 1),
            ],
            preconditions=[unvisited.contains(i), vehicles < k],
        )
        model.add_transition(visit_via_depot)

    name = "return"
    name_to_partial_tour[name] = (nodes[0],)
    return_to_depot = dp.Transition(
        name=name,
        cost=dp.IntExpr.state_cost() + distance[location, 0],
        effects=[(location, 0)],
        preconditions=[unvisited.is_empty(), location != 0],
    )
    model.add_transition(return_to_depot)

    model.add_state_constr((k - vehicles + 1) * capacity >= load + demand[unvisited])

    min_distance_to = model.add_int_table(
        [min(distance_matrix[i][j] for i in range(n) if i != j) for j in range(n)]
    )
    model.add_dual_bound(
        min_distance_to[unvisited] + (location != 0).if_then_else(min_distance_to[0], 0)
    )

    min_distance_from = model.add_int_table(
        [min(distance_matrix[i][j] for j in range(n) if i != j) for i in range(n)]
    )
    model.add_dual_bound(
        min_distance_from[unvisited]
        + (location != 0).if_then_else(min_distance_from[location], 0)
    )

    solver = dp.CABS(model, time_limit=time_limit, quiet=False)
    solution = solver.search()

    if solution.is_infeasible:
        print("Problem is infeasible")
    elif solution.is_optimal:
        print("Optimally solved")
    elif solution.time_out:
        print("Time out")

    tour = [1]

    for t in solution.transitions:
        print(t.name)
        tour += list(name_to_partial_tour[t.name])

    print("expanded: {}".format(solution.expanded))
    print("cost: {}".format(solution.cost))

    return tour, solution.cost


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("input", type=str)
    parser.add_argument("--time-out", default=1800, type=int)
    args = parser.parse_args()

    name = os.path.basename(args.input)
    m = re.match(r".+k(\d+).+", name)
    k = int(m.group(1))

    (
        n,
        nodes,
        edges,
        capacity,
        demand,
        depot,
        _,
    ) = read_tsplib.read_cvrp(args.input)
    tour, cost = solve(n, nodes, edges, capacity, demand, k, time_limit=args.time_out)

    if cost is not None and read_tsplib.validate_cvrp(
        n, nodes, edges, capacity, demand, depot, tour, cost, k
    ):
        print("Solution is valid.")
    else:
        print("Solution is invalid.")



#!/usr/bin/env python3

import argparse

import didppy as dp

import read_mosp


def solve(item_to_patterns, pattern_to_items, time_limit=None):
    m = len(item_to_patterns)
    item_to_neighbors = read_mosp.compute_item_to_neighbors(
        item_to_patterns, pattern_to_items
    )

    model = dp.Model()
    item = model.add_object_type(m)
    remaining = model.add_set_var(item, [i for i in range(m)])
    opened = model.add_set_var(item, [])
    neighbors = model.add_set_table(item_to_neighbors, object_type=item)

    model.add_base_case([remaining.is_empty()])

    name_to_item = {}

    for i in range(m):
        name = "close {}".format(i)
        name_to_item[name] = i
        t = dp.Transition(
            name=name,
            cost=dp.max(
                dp.IntExpr.state_cost(),
                ((opened & remaining) | (neighbors[i] - opened)).len(),
            ),
            effects=[(remaining, remaining.remove(i)), (opened, opened | neighbors[i])],
            preconditions=[remaining.contains(i)],
        )
        model.add_transition(t)

    model.add_dual_bound(0)

    solver = dp.CABS(
        model, f_operator=dp.FOperator.Max, time_limit=time_limit, quiet=False
    )
    solution = solver.search()

    if solution.is_infeasible:
        print("Problem is infeasible")
    elif solution.is_optimal:
        print("Optimally solved")
    elif solution.time_out:
        print("Time out")

    pattern_sequence = []
    produced = set()

    for t in solution.transitions:
        print(t.name)
        item = name_to_item[t.name]

        for pattern in item_to_patterns[item]:
            if pattern not in produced:
                pattern_sequence.append(pattern)
                produced.add(pattern)

    print("expanded: {}".format(solution.expanded))
    print("cost: {}".format(solution.cost))

    return pattern_sequence, solution.cost


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("input", type=str)
    parser.add_argument("--time-out", default=1800, type=int)
    args = parser.parse_args()

    item_to_patterns, pattern_to_items = read_mosp.read(args.input)
    sequence, cost = solve(item_to_patterns, pattern_to_items)

    if cost is not None and read_mosp.validate(
        item_to_patterns, pattern_to_items, sequence, cost
    ):
        print("Solution is valid.")
    else:
        print("Solution is invalid.")



import didppy as dp
import pytest


class TestStateElementVariable:
    model = dp.Model()
    obj = model.add_object_type(number=4)
    var1 = model.add_element_var(object_type=obj, target=2)
    var2 = model.add_element_var(object_type=obj, target=1)

    def test_get(self):
        state = self.model.target_state

        assert state[self.var1] == 2
        assert state[self.var2] == 1

    def test_get_error(self):
        model = dp.Model()
        state = model.target_state

        with pytest.raises(BaseException):
            state[self.var1]

    def test_set(self):
        state = self.model.target_state
        state[self.var1] = 3

        assert state[self.var1] == 3
        assert state[self.var2] == 1

    def test_set_error(self):
        state = self.model.target_state

        with pytest.raises(OverflowError):
            state[self.var1] = -1


class TestStateElementResourceVariable:
    model = dp.Model()
    obj = model.add_object_type(number=4)
    var1 = model.add_element_resource_var(
        object_type=obj, target=2, less_is_better=True
    )
    var2 = model.add_element_resource_var(
        object_type=obj, target=1, less_is_better=False
    )

    def test_get(self):
        state = self.model.target_state

        assert state[self.var1] == 2
        assert state[self.var2] == 1

    def test_get_error(self):
        model = dp.Model()
        state = model.target_state

        with pytest.raises(BaseException):
            state[self.var1]

    def test_set(self):
        state = self.model.target_state
        state[self.var1] = 3

        assert state[self.var1] == 3
        assert state[self.var2] == 1

    def test_set_error(self):
        state = self.model.target_state

        with pytest.raises(OverflowError):
            state[self.var1] = -1


class TestStateSetVariable:
    model = dp.Model()
    obj = model.add_object_type(number=4)
    var1 = model.add_set_var(object_type=obj, target=[0, 1])
    var2 = model.add_set_var(object_type=obj, target=[1, 2])

    def test_get(self):
        state = self.model.target_state

        assert state[self.var1] == {0, 1}
        assert state[self.var2] == {1, 2}

    def test_get_error(self):
        model = dp.Model()
        state = model.target_state

        with pytest.raises(BaseException):
            state[self.var1]

    def test_set(self):
        state = self.model.target_state
        state[self.var1] = self.model.create_set_const(
            object_type=self.obj, value=[0, 2]
        )

        assert state[self.var1] == {0, 2}
        assert state[self.var2] == {1, 2}


class TestStateIntVariable:
    model = dp.Model()
    var1 = model.add_int_var(target=2)
    var2 = model.add_int_var(target=1)

    def test_get(self):
        state = self.model.target_state

        assert state[self.var1] == 2
        assert state[self.var2] == 1

    def test_get_error(self):
        model = dp.Model()
        state = model.target_state

        with pytest.raises(BaseException):
            state[self.var1]

    set_cases = [3, -3]

    @pytest.mark.parametrize("value", set_cases)
    def test_set(self, value):
        state = self.model.target_state
        state[self.var1] = value

        assert state[self.var1] == value
        assert state[self.var2] == 1

    def test_set_error(self):
        state = self.model.target_state

        with pytest.raises(TypeError):
            state[self.var1] = 1.5


class TestStateIntResourceVariable:
    model = dp.Model()
    var1 = model.add_int_resource_var(target=2, less_is_better=True)
    var2 = model.add_int_resource_var(target=1, less_is_better=False)

    def test_get(self):
        state = self.model.target_state

        assert state[self.var1] == 2
        assert state[self.var2] == 1

    def test_get_error(self):
        model = dp.Model()
        state = model.target_state

        with pytest.raises(BaseException):
            state[self.var1]

    set_cases = [3, -3]

    @pytest.mark.parametrize("value", set_cases)
    def test_set(self, value):
        state = self.model.target_state
        state[self.var1] = value

        assert state[self.var1] == value
        assert state[self.var2] == 1

    def test_set_error(self):
        state = self.model.target_state

        with pytest.raises(TypeError):
            state[self.var1] = 1.5


class TestStateFloatVariable:
    model = dp.Model()
    var1 = model.add_float_var(target=2.5)
    var2 = model.add_float_var(target=1.5)

    def test_get(self):
        state = self.model.target_state

        assert state[self.var1] == pytest.approx(2.5)
        assert state[self.var2] == pytest.approx(1.5)

    def test_get_error(self):
        model = dp.Model()
        state = model.target_state

        with pytest.raises(BaseException):
            state[self.var1]

    set_cases = [
        (3.5, pytest.approx(3.5)),
        (-3.5, pytest.approx(-3.5)),
        (3, pytest.approx(3.0)),
        (-3, pytest.approx(-3.0)),
    ]

    @pytest.mark.parametrize("value, expected", set_cases)
    def test_set(self, value, expected):
        state = self.model.target_state
        state[self.var1] = value

        assert state[self.var1] == expected
        assert state[self.var2] == pytest.approx(1.5)


class TestStateFloatResourceVariable:
    model = dp.Model()
    var1 = model.add_float_resource_var(target=2.5, less_is_better=True)
    var2 = model.add_float_resource_var(target=1.5, less_is_better=False)

    def test_get(self):
        state = self.model.target_state

        assert state[self.var1] == pytest.approx(2.5)
        assert state[self.var2] == pytest.approx(1.5)

    def test_get_error(self):
        model = dp.Model()
        state = model.target_state

        with pytest.raises(BaseException):
            state[self.var1]

    set_cases = [
        (3.5, pytest.approx(3.5)),
        (-3.5, pytest.approx(-3.5)),
        (3, pytest.approx(3.0)),
        (-3, pytest.approx(-3.0)),
    ]

    @pytest.mark.parametrize("value, expected", set_cases)
    def test_set(self, value, expected):
        state = self.model.target_state
        state[self.var1] = value

        assert state[self.var1] == expected
        assert state[self.var2] == pytest.approx(1.5)


get_error_cases = [
    0,
    0.5,
    dp.ElementExpr(0),
    dp.IntExpr(0),
    dp.FloatExpr(0.5),
]


@pytest.mark.parametrize("value", get_error_cases)
def test_get_variable_error(value):
    model = dp.Model()
    state = model.target_state

    with pytest.raises(TypeError):
        state[value]



def read(input_file):
    with open(input_file) as f:
        line = f.readline()
        pair = line.split()
        n_patterns = int(pair[0])
        n_items = int(pair[1])
        item_to_patterns = [[] for _ in range(n_items)]
        pattern_to_items = [[] for _ in range(n_patterns)]

        for j in range(n_patterns):
            line = f.readline()
            row = [int(c) for c in line.split()]
            for i in range(n_items):
                if row[i] == 1:
                    item_to_patterns[i].append(j)
                    pattern_to_items[j].append(i)

    return item_to_patterns, pattern_to_items


def compute_item_to_neighbors(item_to_patterns, pattern_to_items):
    m = len(item_to_patterns)
    item_to_neighbors = [set() for _ in range(m)]
    for i in range(m):
        item_to_neighbors[i].add(i)
        for j in item_to_patterns[i]:
            for k in pattern_to_items[j]:
                item_to_neighbors[i].add(k)
    return item_to_neighbors


def item_order_to_pattern_order(item_to_patterns, items):
    produced = set()
    solution = []
    for i in items:
        for j in item_to_patterns[i]:
            if j not in produced:
                solution.append(j)
                produced.add(j)
    return solution


def validate(item_to_patterns, pattern_to_items, solution, cost):
    actual_cost = 0
    produced = set()
    open = set()
    for i in solution:
        if i < 0 or i > len(pattern_to_items) - 1:
            print("Pattern {} does not exist".format(i))
            return False
        if i in produced:
            print("Pattern {} is already produced".format(i))
            return False

        produced.add(i)
        for j in pattern_to_items[i]:
            if j not in open:
                open.add(j)
        actual_cost = max(actual_cost, len(open))
        closed = []
        for j in open:
            if all([k in produced for k in item_to_patterns[j]]):
                closed.append(j)
        for j in closed:
            open.remove(j)

    if len(produced) != len(pattern_to_items):
        print(
            "The number of produced patterns is {}, but should be {}".format(
                len(produced), len(pattern_to_items)
            )
        )
        return False

    if cost != actual_cost:
        print(
            "The cost of the solution {} mismatches the actual cost {}".format(
                cost, actual_cost
            )
        )
        return False

    return True



def read(filename):
    with open(filename) as f:
        line = f.readline().rstrip().split()
        n = int(line[0])
        line = f.readline().rstrip().split()
        node_weights = [int(i) for i in line]
        edge_weights = {}
        for i in range(n):
            line = [int(w) for w in f.readline().rstrip().split()]
            for j, w in enumerate(line):
                if j > i and w > 0:
                    edge_weights[i, j] = w
        return n, node_weights, edge_weights


def validate(n, node_weights, edge_weights, solution, cost):
    actual_cost = 0
    clean = set()

    for i in solution:
        if i < 0 or i > n - 1:
            print("Node {} does not exist".format(i))
            return False
        if i in clean:
            print("{} is already clean".format(i))
            return False

        n_robots = node_weights[i]
        for j in range(n):
            if (i, j) in edge_weights:
                n_robots += edge_weights[i, j]
            elif (j, i) in edge_weights:
                n_robots += edge_weights[j, i]
        for j in range(n):
            if j in clean:
                for k in range(n):
                    if k != i and k not in clean:
                        if (j, k) in edge_weights:
                            n_robots += edge_weights[j, k]
                        elif (k, j) in edge_weights:
                            n_robots += edge_weights[k, j]

        actual_cost = max(actual_cost, n_robots)
        clean.add(i)

    if len(clean) != n:
        print("The number of swept nodes is {}, but should be {}".format(len(clean), n))
        return False

    if actual_cost != cost:
        print(
            "The cost of solution {} mismatches the actual cost {}".format(
                cost, actual_cost
            )
        )
        return False

    return True



#!/usr/bin/env python3

import argparse

import didppy as dp

import read_graph_clear


def solve(n, node_weights, edge_weights, time_limit=None):
    model = dp.Model()
    node = model.add_object_type(n)
    clean = model.add_set_var(node, [])
    all_nodes = model.create_set_const(node, [i for i in range(n)])
    a = model.add_int_table(node_weights)
    b = model.add_int_table(
        [
            [
                edge_weights[i, j]
                if (i, j) in edge_weights
                else edge_weights[j, i]
                if (j, i) in edge_weights
                else 0
                for j in range(n)
            ]
            for i in range(n)
        ]
    )

    model.add_base_case([all_nodes <= clean])

    name_to_node = {}

    for i in range(n):
        name = "sweep {}".format(i)
        name_to_node[name] = i
        t = dp.Transition(
            name=name,
            cost=dp.max(
                dp.IntExpr.state_cost(),
                a[i] + b[i, all_nodes] + b[clean, clean.complement().remove(i)],
            ),
            effects=[(clean, clean.add(i))],
            preconditions=[~clean.contains(i)],
        )
        model.add_transition(t)

    model.add_dual_bound(0)

    solver = dp.CABS(
        model, f_operator=dp.FOperator.Max, time_limit=time_limit, quiet=False
    )
    solution = solver.search()

    if solution.is_infeasible:
        print("Problem is infeasible")
    elif solution.is_optimal:
        print("Optimally solved")
    elif solution.time_out:
        print("Time out")

    sequence = []

    for t in solution.transitions:
        print(t.name)
        sequence.append(name_to_node[t.name])

    print("expanded: {}".format(solution.expanded))
    print("cost: {}".format(solution.cost))

    return sequence, solution.cost


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("input", type=str)
    parser.add_argument("--time-out", default=1800, type=int)
    args = parser.parse_args()

    n, a, b = read_graph_clear.read(args.input)
    sequence, cost = solve(n, a, b, time_limit=args.time_out)

    if cost is not None and read_graph_clear.validate(n, a, b, sequence, cost):
        print("Solution is valid.")
    else:
        print("Solution is invalid.")



#! /usr/bin/env python3

import argparse
import math

import didppy as dp

import read_salbp1


def solve(
    number_of_tasks,
    cycle_time,
    task_times,
    predecessors,
    time_limit=None,
):
    model = dp.Model()
    task = model.add_object_type(number=number_of_tasks)
    uncompleted = model.add_set_var(
        object_type=task, target=list(range(number_of_tasks))
    )
    idle_time = model.add_int_resource_var(target=0, less_is_better=False)
    task_time_table = model.add_int_table(
        [task_times[i + 1] for i in range(number_of_tasks)]
    )
    predecessors_table = model.add_set_table(
        [[j - 1 for j in predecessors[i + 1]] for i in range(number_of_tasks)],
        object_type=task,
    )
    lb2_weight1 = model.add_int_table(
        [1 if task_times[i + 1] > cycle_time / 2 else 0 for i in range(number_of_tasks)]
    )
    lb2_weight2 = model.add_float_table(
        [
            0.5 if task_times[i + 1] == cycle_time / 2 else 0
            for i in range(number_of_tasks)
        ]
    )
    lb3_weight = model.add_float_table(
        [
            1.0
            if task_times[i + 1] > cycle_time * 2 / 3
            else 2 / 3 // 0.001 / 1000
            if task_times[i + 1] == cycle_time * 2 / 3
            else 0.5
            if task_times[i + 1] > cycle_time / 3
            else 1 / 3 // 0.001 / 1000
            if task_times[i + 1] == cycle_time / 3
            else 0.0
            for i in range(number_of_tasks)
        ]
    )
    model.add_base_case([uncompleted.is_empty()])

    name_to_task = {}

    for i in range(number_of_tasks):
        name = "schedule {}".format(i)
        name_to_task[name] = i + 1
        t = dp.Transition(
            name=name,
            cost=dp.IntExpr.state_cost(),
            effects=[
                (uncompleted, uncompleted.remove(i)),
                (idle_time, idle_time - task_time_table[i]),
            ],
            preconditions=[
                uncompleted.contains(i),
                task_time_table[i] <= idle_time,
                (uncompleted & predecessors_table[i]).is_empty(),
            ],
        )
        model.add_transition(t)

    t = dp.Transition(
        name="open a new station",
        cost=dp.IntExpr.state_cost() + 1,
        effects=[(idle_time, cycle_time)],
        preconditions=[
            ~uncompleted.contains(i)
            | (task_time_table[i] > idle_time)
            | ((uncompleted & predecessors_table[i]).len() > 0)
            for i in range(number_of_tasks)
        ],
    )
    model.add_transition(t, forced=True)

    model.add_dual_bound(
        math.ceil((task_time_table[uncompleted] - idle_time) / cycle_time)
    )
    model.add_dual_bound(
        lb2_weight1[uncompleted]
        + math.ceil(lb2_weight2[uncompleted])
        - (idle_time >= cycle_time / 2).if_then_else(1, 0)
    )
    model.add_dual_bound(
        math.ceil(lb3_weight[uncompleted])
        - (idle_time >= cycle_time / 3).if_then_else(1, 0)
    )

    solver = dp.CABS(model, time_limit=time_limit, quiet=False)
    solution = solver.search()

    if solution.is_infeasible:
        print("Problem is infeasible")
    elif solution.is_optimal:
        print("Optimally solved")
    elif solution.time_out:
        print("Time out")

    assignment = []

    for t in solution.transitions:
        print(t.name)

        if t.name == "open a new station":
            assignment.append([])
        else:
            assignment[-1].append(name_to_task[t.name])

    print("expanded: {}".format(solution.expanded))
    print("cost: {}".format(solution.cost))

    return assignment, solution.cost


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("input", type=str)
    parser.add_argument("--time-out", default=1800, type=int)
    args = parser.parse_args()

    number_of_tasks, cycle_time, task_times, predecessors, _ = read_salbp1.read(
        args.input
    )
    assignment, cost = solve(
        number_of_tasks,
        cycle_time,
        task_times,
        predecessors,
        time_limit=args.time_out,
    )
    if cost is not None and read_salbp1.validate(
        number_of_tasks, cycle_time, task_times, predecessors, assignment, cost
    ):
        print("Solution is valid.")
    else:
        print("Solution is invalid.")



import math


def tsplib_round(x):
    return math.floor(x + 0.5)


def read_euc2d(n, f):
    line = f.readline()
    while True:
        if line.startswith("NODE_COORD_SECTION"):
            break
        line = f.readline()
    nodes = []
    edges = {}
    x = {}
    y = {}
    for _ in range(n):
        row = f.readline().split()
        i = int(row[0])
        nodes.append(i)
        x[i] = float(row[1])
        y[i] = float(row[2])
    for i in nodes:
        for j in nodes:
            if i >= j:
                continue
            edges[i, j] = tsplib_round(
                math.sqrt((x[i] - x[j]) ** 2 + (y[i] - y[j]) ** 2)
            )
            edges[j, i] = edges[i, j]
    return nodes, edges, True


def read_euc3d(n, f):
    line = f.readline()
    while True:
        if line.startswith("NODE_COORD_SECTION"):
            break
        line = f.readline()
    nodes = []
    edges = {}
    x = {}
    y = {}
    z = {}
    for _ in range(n):
        row = f.readline().split()
        i = int(row[0])
        nodes.append(i)
        x[i] = float(row[1])
        y[i] = float(row[2])
        z[i] = float(row[3])
    for i in nodes:
        for j in nodes:
            if i >= j:
                continue
            edges[i, j] = tsplib_round(
                math.sqrt((x[i] - x[j]) ** 2 + (y[i] - y[j]) ** 2 + (z[i] - z[j]) ** 2)
            )
            edges[j, i] = edges[i, j]
    return nodes, edges, True


def read_geo(n, f):
    line = f.readline()
    while True:
        if line.startswith("NODE_COORD_SECTION"):
            break
        line = f.readline()
    PI = 3.141592
    nodes = []
    edges = {}
    latitude = {}
    longtitude = {}
    for _ in range(n):
        row = f.readline().split()
        i = int(row[0])
        nodes.append(i)
        x = float(row[1])
        deg = int(x)
        min = x - deg
        latitude[i] = PI * (deg + 5.0 * min / 3.0) / 180.0
        y = float(row[2])
        deg = int(y)
        min = y - deg
        longtitude[i] = PI * (deg + 5.0 * min / 3.0) / 180.0
    RRR = 6378.388
    for i in nodes:
        for j in nodes:
            if i >= j:
                continue
            q1 = math.cos(longtitude[i] - longtitude[j])
            q2 = math.cos(latitude[i] - latitude[j])
            q3 = math.cos(latitude[i] + latitude[j])
            edges[i, j] = int(
                RRR * math.acos(0.5 * ((1.0 + q1) * q2 - (1.0 - q1) * q3)) + 1.0
            )
            edges[j, i] = edges[i, j]
    return nodes, edges, True


def read_att(n, f):
    line = f.readline()
    while True:
        if line.startswith("NODE_COORD_SECTION"):
            break
        line = f.readline()
    nodes = []
    edges = {}
    x = {}
    y = {}
    for _ in range(n):
        row = f.readline().split()
        i = int(row[0])
        nodes.append(i)
        x[i] = float(row[1])
        y[i] = float(row[2])
    for i in nodes:
        for j in nodes:
            if i >= j:
                continue
            xd = x[i] - x[j]
            yd = y[i] - y[j]
            rij = math.sqrt((xd**2 + yd**2) / 10.0)
            tij = tsplib_round(rij)
            if tij < rij:
                edges[i, j] = tij + 1
            else:
                edges[i, j] = tij
            edges[j, i] = edges[i, j]
    return nodes, edges, True


def read_ceil2d(n, f):
    line = f.readline()
    while True:
        if line.startswith("NODE_COORD_SECTION"):
            break
        line = f.readline()
    nodes = []
    edges = {}
    x = {}
    y = {}
    for _ in range(n):
        row = f.readline().split()
        i = int(row[0])
        nodes.append(i)
        x[i] = float(row[1])
        y[i] = float(row[2])
    for i in nodes:
        for j in nodes:
            if i >= j:
                continue
            edges[i, j] = math.ceil(math.sqrt((x[i] - x[j]) ** 2 + (y[i] - y[j]) ** 2))
            edges[j, i] = edges[i, j]
    return nodes, edges, True


def read_full_matrix(n, f):
    line = f.readline()
    while True:
        if line.startswith("EDGE_WEIGHT_SECTION"):
            break
        line = f.readline()
    edges = {}
    line = f.readline()
    entries = [int(c) for c in line.split()]
    position = 0
    symmetric = True
    for i in range(1, n + 1):
        for j in range(1, n + 1):
            if position >= len(entries):
                line = f.readline()
                entries = [int(c) for c in line.split()]
                position = 0
            edges[i, j] = entries[position]
            position += 1
    for i in range(1, n):
        for j in range(i + 1, n + 1):
            if edges[i, j] != edges[j, i]:
                symmetric = False
                break
    return edges, symmetric


def read_upper_row(n, f):
    line = f.readline()
    while True:
        if line.startswith("EDGE_WEIGHT_SECTION"):
            break
        line = f.readline()
    edges = {}
    line = f.readline()
    entries = [int(c) for c in line.split()]
    position = 0
    for i in range(1, n):
        for j in range(i + 1, n + 1):
            if position >= len(entries):
                line = f.readline()
                entries = [int(c) for c in line.split()]
                position = 0
            edges[i, j] = entries[position]
            edges[j, i] = edges[i, j]
            position += 1
    return edges, True


def read_lower_row(n, f):
    line = f.readline()
    while True:
        if line.startswith("EDGE_WEIGHT_SECTION"):
            break
        line = f.readline()
    edges = {}
    line = f.readline()
    entries = [int(c) for c in line.split()]
    position = 0
    for i in range(2, n + 1):
        for j in range(1, i):
            if position >= len(entries):
                line = f.readline()
                entries = [int(c) for c in line.split()]
                position = 0
            edges[i, j] = entries[position]
            edges[j, i] = edges[i, j]
            position += 1
    return edges, True


def read_upper_diag_row(n, f):
    line = f.readline()
    while True:
        if line.startswith("EDGE_WEIGHT_SECTION"):
            break
        line = f.readline()
    edges = {}
    line = f.readline()
    entries = [int(c) for c in line.split()]
    position = 0
    for i in range(1, n + 1):
        for j in range(i, n + 1):
            if position >= len(entries):
                line = f.readline()
                entries = [int(c) for c in line.split()]
                position = 0
            edges[i, j] = entries[position]
            if i != j:
                edges[j, i] = edges[i, j]
            position += 1
    return edges, True


def read_lower_diag_row(n, f):
    line = f.readline()
    while line:
        if line.startswith("EDGE_WEIGHT_SECTION"):
            break
        line = f.readline()
    entries = []
    line = f.readline()
    edges = {}
    line = f.readline()
    entries = [int(c) for c in line.split()]
    position = 0
    for i in range(1, n + 1):
        for j in range(1, i + 1):
            if position >= len(entries):
                line = f.readline()
                entries = [int(c) for c in line.split()]
                position = 0
            edges[i, j] = entries[position]
            if i != j:
                edges[j, i] = edges[i, j]
            position += 1
    return edges, True


def read_edges(edge_weight_type, edge_weight_format, n, f):
    if edge_weight_type == "EUC_2D":
        return read_euc2d(n, f)

    if edge_weight_type == "EUC_3D":
        return read_euc3d(n, f)

    if edge_weight_type == "GEO":
        return read_geo(n, f)

    if edge_weight_type == "ATT":
        return read_att(n, f)

    if edge_weight_type == "CEIL_2D":
        return read_ceil2d(n, f)

    nodes = list(range(1, n + 1))

    if edge_weight_format == "FULL_MATRIX":
        return nodes, *read_full_matrix(n, f)

    if edge_weight_format == "UPPER_ROW":
        return nodes, *read_upper_row(n, f)

    if edge_weight_format == "LOWER_ROW":
        return nodes, *read_lower_row(n, f)

    if edge_weight_format == "UPPER_DIAG_ROW":
        return nodes, *read_upper_diag_row(n, f)

    if edge_weight_format == "LOWER_DIAG_ROW":
        return nodes, *read_lower_diag_row(n, f)

    return None, None, None


def read_tsp(filename):
    with open(filename) as f:
        n = None
        edge_weight_type = None
        edge_weight_format = None
        while (
            n is None
            or edge_weight_type is None
            or (edge_weight_type == "EXPLICIT" and edge_weight_format is None)
        ):
            line = f.readline()
            if line.startswith("TYPE"):
                assert "TSP" in line.split() or "ATSP" in line.split()
            if line.startswith("DIMENSION"):
                n = int(line.split()[-1])
            if line.startswith("EDGE_WEIGHT_TYPE"):
                edge_weight_type = line.split()[-1]
            if line.startswith("EDGE_WEIGHT_FORMAT"):
                edge_weight_format = line.split()[-1]

        nodes, edges, symmetric = read_edges(edge_weight_type, edge_weight_format, n, f)
        return n, nodes, edges, symmetric


def read_demand(n, f):
    line = f.readline()
    while line:
        if line.startswith("DEMAND_SECTION"):
            break
        line = f.readline()

    demand = {}

    for _ in range(n):
        line = f.readline().split()
        demand[int(line[0])] = int(line[1])

    return demand


def read_depots(n, f):
    line = f.readline()
    while line:
        if line.startswith("DEPOT_SECTION"):
            break
        line = f.readline()

    depots = []
    d = int(f.readline().rstrip())
    while d != -1:
        depots.append(d)
        d = int(f.readline().rstrip())

    return depots


def read_cvrp(filename):
    with open(filename) as f:
        n = None
        edge_weight_type = None
        edge_weight_format = None
        capacity = None
        while (
            n is None
            or edge_weight_type is None
            or capacity is None
            or (edge_weight_type == "EXPLICIT" and edge_weight_format is None)
        ):
            line = f.readline()
            if line.startswith("TYPE"):
                assert "CVRP" in line.split()
            if line.startswith("DIMENSION"):
                n = int(line.split()[-1])
            if line.startswith("EDGE_WEIGHT_TYPE"):
                edge_weight_type = line.split()[-1]
            if line.startswith("CAPACITY"):
                capacity = int(line.split()[-1])
            if line.startswith("EDGE_WEIGHT_FORMAT"):
                edge_weight_format = line.split()[-1]

        nodes, edges, symmetric = read_edges(edge_weight_type, edge_weight_format, n, f)
        demand = read_demand(n, f)
        depots = read_depots(n, f)
        assert len(depots) == 1
        return n, nodes, edges, capacity, demand, depots[0], symmetric


def validate_cvrp(n, nodes, edges, capacity, demand, depot, solution, cost, k=None):
    if solution[0] != depot:
        print(
            "The tour does not start from the depot {} but from {}".format(
                depot, solution[0]
            )
        )
        return False
    previous = solution[0]
    actual_cost = 0
    load = 0
    n_vehicles = 0
    visited = set([depot])
    for i in solution[1:]:
        if previous == depot:
            n_vehicles += 1
            if k is not None and n_vehicles > k:
                print(
                    "The number of vehicles {} exceeds the limit {}".format(
                        n_vehicles, k
                    )
                )
                return False

        if i not in nodes:
            print("No such customer {}".format(i))
            return False

        if (previous, i) not in edges:
            print("No such edge ({}, {})".format(previous, i))
            return False

        actual_cost += edges[previous, i]

        if i == depot:
            load = 0
        else:
            if i in visited:
                print("Customer {} is already visited".format(i))
                return False
            visited.add(i)
            load += demand[i]
            if load > capacity:
                print(
                    "load {} at customer {} exceeds the capacity {}".format(
                        load, i, capacity
                    )
                )
                return False

        previous = i

    if previous != depot:
        print(
            "The tour does not return to the depot {} but to {}".format(depot, previous)
        )
        return False

    if len(visited) != n:
        print(
            "The number of visited customers is {}, but should be {}".format(
                len(visited), n
            )
        )

    if actual_cost != cost:
        print(
            "The cost of the solution {} mismatches the actual cost {}".format(
                cost, actual_cost
            )
        )
        return False

    return True



import didppy as dp
import pytest


class TestElementTable1D:
    model = dp.Model()
    obj = model.add_object_type(number=2)

    zero_expr = dp.ElementExpr(0)
    zero_var = model.add_element_var(object_type=obj, target=0)
    zero_resource_var = model.add_element_resource_var(
        object_type=obj, target=0, less_is_better=True
    )

    one_expr = dp.ElementExpr(1)
    one_var = model.add_element_var(object_type=obj, target=1)
    one_resource_var = model.add_element_resource_var(
        object_type=obj, target=1, less_is_better=True
    )

    table = model.add_element_table([1, 0])
    state = model.target_state

    cases = [
        (zero_expr, 1),
        (zero_var, 1),
        (zero_resource_var, 1),
        (0, 1),
        (one_expr, 0),
        (one_var, 0),
        (one_resource_var, 0),
        (1, 0),
    ]

    @pytest.mark.parametrize("x, expected", cases)
    def test(self, x, expected):
        assert self.table[x].eval(self.state, self.model) == expected

    error_cases = [-1, dp.IntExpr(0), 2]

    @pytest.mark.parametrize("index", error_cases)
    def test_error(self, index):
        with pytest.raises(BaseException):
            self.table[index].eval(self.state, self.model)


class TestElementTable2D:
    model = dp.Model()
    obj = model.add_object_type(number=2)

    zero_expr = dp.ElementExpr(0)
    zero_var = model.add_element_var(object_type=obj, target=0)
    zero_resource_var = model.add_element_resource_var(
        object_type=obj, target=0, less_is_better=True
    )

    one_expr = dp.ElementExpr(1)
    one_var = model.add_element_var(object_type=obj, target=1)
    one_resource_var = model.add_element_resource_var(
        object_type=obj, target=1, less_is_better=True
    )

    table = model.add_element_table([[1, 0], [0, 1]])
    state = model.target_state

    cases = [
        (zero, zero, 1) for zero in [zero_expr, zero_var, zero_resource_var, 0]
    ] + [(one, one, 1) for one in [one_expr, one_var, one_resource_var, 1]]

    for zero in [zero_expr, zero_var, zero_resource_var, 0]:
        for one in [one_expr, one_var, one_resource_var, 1]:
            cases.append((zero, one, 0))
            cases.append((one, zero, 0))

    @pytest.mark.parametrize("x, y, expected", cases)
    def test(self, x, y, expected):
        assert self.table[x, y].eval(self.state, self.model) == expected

    error_cases = [
        0,
        (0, -1),
        (-1, 0),
        (0, dp.IntExpr(1)),
        (dp.IntExpr(1), 0),
        (0, 2),
        (2, 0),
    ]

    @pytest.mark.parametrize("index", error_cases)
    def test_error(self, index):
        with pytest.raises(BaseException):
            self.table[index].eval(self.state, self.model)


class TestElementTable3D:
    model = dp.Model()
    obj = model.add_object_type(number=3)

    zero_expr = dp.ElementExpr(0)
    zero_var = model.add_element_var(object_type=obj, target=0)
    zero_resource_var = model.add_element_resource_var(
        object_type=obj, target=0, less_is_better=True
    )

    one_expr = dp.ElementExpr(1)
    one_var = model.add_element_var(object_type=obj, target=1)
    one_resource_var = model.add_element_resource_var(
        object_type=obj, target=1, less_is_better=True
    )

    two_expr = dp.ElementExpr(2)
    two_var = model.add_element_var(object_type=obj, target=2)
    two_resource_var = model.add_element_resource_var(
        object_type=obj, target=2, less_is_better=True
    )

    table = model.add_element_table(
        [
            [[1, 0, 2], [2, 0, 1], [1, 2, 0]],
            [[0, 2, 1], [0, 1, 2], [1, 0, 2]],
            [[1, 2, 0], [0, 1, 2], [0, 2, 1]],
        ]
    )
    state = model.target_state

    cases = (
        [(zero, zero, zero, 1) for zero in [zero_expr, zero_var, zero_resource_var, 0]]
        + [(one, one, one, 1) for one in [one_expr, one_var, one_resource_var, 1]]
        + [(two, two, two, 1) for two in [two_expr, two_var, two_resource_var, 2]]
    )

    for zero in [zero_expr, zero_var, zero_resource_var, 0]:
        for one in [one_expr, one_var, one_resource_var, 1]:
            for two in [two_expr, two_var, two_resource_var, 2]:
                cases.append((zero, one, two, 1))
                cases.append((zero, two, one, 2))
                cases.append((one, zero, two, 1))
                cases.append((one, two, zero, 1))
                cases.append((two, zero, one, 2))
                cases.append((two, one, zero, 0))

    for zero in [zero_expr, zero_var, zero_resource_var, 0]:
        for one in [one_expr, one_var, one_resource_var, 1]:
            cases.append((zero, zero, one, 0))
            cases.append((zero, one, zero, 2))
            cases.append((zero, one, one, 0))
            cases.append((one, zero, zero, 0))
            cases.append((one, zero, one, 2))
            cases.append((one, one, zero, 0))

    for zero in [zero_expr, zero_var, zero_resource_var, 0]:
        for two in [two_expr, two_var, two_resource_var, 2]:
            cases.append((zero, zero, two, 2))
            cases.append((zero, two, zero, 1))
            cases.append((zero, two, two, 0))
            cases.append((two, zero, zero, 1))
            cases.append((two, zero, two, 0))
            cases.append((two, two, zero, 0))

    for one in [one_expr, one_var, one_resource_var, 1]:
        for two in [two_expr, two_var, two_resource_var, 2]:
            cases.append((one, one, two, 2))
            cases.append((one, two, one, 0))
            cases.append((one, two, two, 2))
            cases.append((two, one, one, 1))
            cases.append((two, one, two, 2))
            cases.append((two, two, one, 2))

    @pytest.mark.parametrize("x, y, z, expected", cases)
    def test(self, x, y, z, expected):
        assert self.table[x, y, z].eval(self.state, self.model) == expected

    error_cases = [
        0,
        (0, 1),
        (-1, 0, 0),
        (0, -1, 0),
        (0, 0, -1),
        (dp.IntExpr(1), 0, 0),
        (0, dp.IntExpr(1), 0),
        (0, 0, dp.IntExpr(1)),
        (3, 0, 0),
        (0, 3, 0),
        (0, 0, 3),
    ]

    @pytest.mark.parametrize("index", error_cases)
    def test_error(self, index):
        with pytest.raises(BaseException):
            self.table[index].eval(self.state, self.model)


class TestElementTable:
    model = dp.Model()
    obj = model.add_object_type(number=3)

    zero_expr = dp.ElementExpr(0)
    zero_var = model.add_element_var(object_type=obj, target=0)
    zero_resource_var = model.add_element_resource_var(
        object_type=obj, target=0, less_is_better=True
    )

    one_expr = dp.ElementExpr(1)
    one_var = model.add_element_var(object_type=obj, target=1)
    one_resource_var = model.add_element_resource_var(
        object_type=obj, target=1, less_is_better=True
    )

    table = model.add_element_table({(0, 0, 0, 0): 1, (1, 1, 1, 1): 0}, default=2)
    state = model.target_state

    cases = (
        [
            ((zero, zero, zero, zero), 1)
            for zero in [zero_expr, zero_var, zero_resource_var, 0]
        ]
        + [
            ((one, one, one, one), 0)
            for one in [one_expr, one_var, one_resource_var, 1]
        ]
        + [((0, 1, 0, 1), 2)]
    )

    @pytest.mark.parametrize("index, expected", cases)
    def test(self, index, expected):
        assert self.table[index].eval(self.state, self.model) == expected

    error_cases = [
        (-1, 0, 0),
        (0, -1, 0),
        (0, 0, -1),
        (dp.IntExpr(1), 0, 0),
        (0, dp.IntExpr(1), 0),
        (0, 0, dp.IntExpr(1)),
    ]

    @pytest.mark.parametrize("index", error_cases)
    def test_error(self, index):
        with pytest.raises(BaseException):
            self.table[index].eval(self.state, self.model)


class TestSetTable1D:
    model = dp.Model()
    obj = model.add_object_type(number=2)

    zero_expr = dp.ElementExpr(0)
    zero_var = model.add_element_var(object_type=obj, target=0)
    zero_resource_var = model.add_element_resource_var(
        object_type=obj, target=0, less_is_better=True
    )

    one_expr = dp.ElementExpr(1)
    one_var = model.add_element_var(object_type=obj, target=1)
    one_resource_var = model.add_element_resource_var(
        object_type=obj, target=1, less_is_better=True
    )

    set_const = model.create_set_const(object_type=obj, value=[0, 1])
    set_expr = dp.SetExpr(set_const)
    set_var = model.add_set_var(object_type=obj, target=[0, 1])

    table = model.add_set_table([[1], [0]], object_type=obj)
    state = model.target_state

    cases = [
        (zero_expr, {1}),
        (zero_var, {1}),
        (zero_resource_var, {1}),
        (0, {1}),
        (one_expr, {0}),
        (one_var, {0}),
        (one_resource_var, {0}),
        (1, {0}),
    ]

    @pytest.mark.parametrize("x, expected", cases)
    def test(self, x, expected):
        assert self.table[x].eval(self.state, self.model) == expected

    union_cases = [
        (set_const, {0, 1}),
        (set_expr, {0, 1}),
        (set_var, {0, 1}),
    ]

    @pytest.mark.parametrize("x, expected", union_cases)
    def test_union(self, x, expected):
        assert self.table.union(x).eval(self.state, self.model) == expected

    intersection_cases = [
        (set_const, set()),
        (set_expr, set()),
        (set_var, set()),
    ]

    @pytest.mark.parametrize("x, expected", intersection_cases)
    def test_intersection(self, x, expected):
        assert self.table.intersection(x).eval(self.state, self.model) == expected

    symmetric_difference_cases = [
        (set_const, {0, 1}),
        (set_expr, {0, 1}),
        (set_var, {0, 1}),
    ]

    @pytest.mark.parametrize("x, expected", symmetric_difference_cases)
    def test_symmetric_difference(self, x, expected):
        assert (
            self.table.symmetric_difference(x).eval(self.state, self.model) == expected
        )

    error_cases = [-1, dp.IntExpr(0), 2]

    @pytest.mark.parametrize("index", error_cases)
    def test_error(self, index):
        with pytest.raises(BaseException):
            self.table[index].eval(self.state, self.model)


class TestSetTable2D:
    model = dp.Model()
    obj = model.add_object_type(number=2)

    zero_expr = dp.ElementExpr(0)
    zero_var = model.add_element_var(object_type=obj, target=0)
    zero_resource_var = model.add_element_resource_var(
        object_type=obj, target=0, less_is_better=True
    )

    one_expr = dp.ElementExpr(1)
    one_var = model.add_element_var(object_type=obj, target=1)
    one_resource_var = model.add_element_resource_var(
        object_type=obj, target=1, less_is_better=True
    )

    set_const = model.create_set_const(object_type=obj, value=[0, 1])
    set_expr = dp.SetExpr(set_const)
    set_var = model.add_set_var(object_type=obj, target=[0, 1])

    table = model.add_set_table([[[1], [0]], [[0], [1]]], object_type=obj)
    state = model.target_state

    cases = [
        (zero, zero, {1}) for zero in [zero_expr, zero_var, zero_resource_var, 0]
    ] + [(one, one, {1}) for one in [one_expr, one_var, one_resource_var, 1]]

    for zero in [zero_expr, zero_var, zero_resource_var, 0]:
        for one in [one_expr, one_var, one_resource_var, 1]:
            cases.append((zero, one, {0}))
            cases.append((one, zero, {0}))

    @pytest.mark.parametrize("x, y, expected", cases)
    def test(self, x, y, expected):
        assert self.table[x, y].eval(self.state, self.model) == expected

    union_cases = []

    for x in [set_const, set_expr, set_var]:
        for y in [set_const, set_expr, set_var]:
            union_cases.append((x, y, {0, 1}))

    for value in [set_const, set_expr, set_var]:
        for zero in [zero_expr, zero_var, zero_resource_var, 0]:
            union_cases.append((value, zero, {0, 1}))
            union_cases.append((zero, value, {0, 1}))

    @pytest.mark.parametrize("x, y, expected", union_cases)
    def test_union(self, x, y, expected):
        assert self.table.union(x, y).eval(self.state, self.model) == expected

    intersection_cases = []

    for x in [set_const, set_expr, set_var]:
        for y in [set_const, set_expr, set_var]:
            intersection_cases.append((x, y, set()))

    for value in [set_const, set_expr, set_var]:
        for zero in [zero_expr, zero_var, zero_resource_var, 0]:
            intersection_cases.append((value, zero, set()))
            intersection_cases.append((zero, value, set()))

    @pytest.mark.parametrize("x, y, expected", intersection_cases)
    def test_intersection(self, x, y, expected):
        assert self.table.intersection(x, y).eval(self.state, self.model) == expected

    symmetric_difference_cases = []

    for x in [set_const, set_expr, set_var]:
        for y in [set_const, set_expr, set_var]:
            symmetric_difference_cases.append((x, y, set()))

    for value in [set_const, set_expr, set_var]:
        for zero in [zero_expr, zero_var, zero_resource_var, 0]:
            symmetric_difference_cases.append((value, zero, {0, 1}))
            symmetric_difference_cases.append((zero, value, {0, 1}))

    @pytest.mark.parametrize("x, y, expected", symmetric_difference_cases)
    def test_symmetric_difference(self, x, y, expected):
        assert (
            self.table.symmetric_difference(x, y).eval(self.state, self.model)
            == expected
        )

    error_cases = [
        0,
        (-1, 0),
        (0, -1),
        (dp.IntExpr(0), 0),
        (0, dp.IntExpr(2)),
        (2, 0),
        (0, 2),
    ]

    @pytest.mark.parametrize("index", error_cases)
    def test_error(self, index):
        with pytest.raises(BaseException):
            self.table[index].eval(self.state, self.model)


class TestSetTable3D:
    model = dp.Model()
    obj = model.add_object_type(number=3)

    zero_expr = dp.ElementExpr(0)
    zero_var = model.add_element_var(object_type=obj, target=0)
    zero_resource_var = model.add_element_resource_var(
        object_type=obj, target=0, less_is_better=True
    )

    one_expr = dp.ElementExpr(1)
    one_var = model.add_element_var(object_type=obj, target=1)
    one_resource_var = model.add_element_resource_var(
        object_type=obj, target=1, less_is_better=True
    )

    two_expr = dp.ElementExpr(2)
    two_var = model.add_element_var(object_type=obj, target=2)
    two_resource_var = model.add_element_resource_var(
        object_type=obj, target=2, less_is_better=True
    )

    set_const = model.create_set_const(object_type=obj, value=[0, 1])
    set_expr = dp.SetExpr(set_const)
    set_var = model.add_set_var(object_type=obj, target=[0, 1])

    table = model.add_set_table(
        [
            [[[1], [0], [2]], [[2], [0], [1]], [[1], [2], [0]]],
            [[[0], [2], [1]], [[0], [1], [2]], [[1], [0], [2]]],
            [[[1], [2], [0]], [[0], [1], [2]], [[0], [2], [1]]],
        ],
        object_type=obj,
    )
    state = model.target_state

    cases = (
        [
            (zero, zero, zero, {1})
            for zero in [zero_expr, zero_var, zero_resource_var, 0]
        ]
        + [(one, one, one, {1}) for one in [one_expr, one_var, one_resource_var, 1]]
        + [(two, two, two, {1}) for two in [two_expr, two_var, two_resource_var, 2]]
    )

    for zero in [zero_expr, zero_var, zero_resource_var, 0]:
        for one in [one_expr, one_var, one_resource_var, 1]:
            for two in [two_expr, two_var, two_resource_var, 2]:
                cases.append((zero, one, two, {1}))
                cases.append((zero, two, one, {2}))
                cases.append((one, zero, two, {1}))
                cases.append((one, two, zero, {1}))
                cases.append((two, zero, one, {2}))
                cases.append((two, one, zero, {0}))

    for zero in [zero_expr, zero_var, zero_resource_var, 0]:
        for one in [one_expr, one_var, one_resource_var, 1]:
            cases.append((zero, zero, one, {0}))
            cases.append((zero, one, zero, {2}))
            cases.append((zero, one, one, {0}))
            cases.append((one, zero, zero, {0}))
            cases.append((one, zero, one, {2}))
            cases.append((one, one, zero, {0}))

    for zero in [zero_expr, zero_var, zero_resource_var, 0]:
        for two in [two_expr, two_var, two_resource_var, 2]:
            cases.append((zero, zero, two, {2}))
            cases.append((zero, two, zero, {1}))
            cases.append((zero, two, two, {0}))
            cases.append((two, zero, zero, {1}))
            cases.append((two, zero, two, {0}))
            cases.append((two, two, zero, {0}))

    for one in [one_expr, one_var, one_resource_var, 1]:
        for two in [two_expr, two_var, two_resource_var, 2]:
            cases.append((one, one, two, {2}))
            cases.append((one, two, one, {0}))
            cases.append((one, two, two, {2}))
            cases.append((two, one, one, {1}))
            cases.append((two, one, two, {2}))
            cases.append((two, two, one, {2}))

    @pytest.mark.parametrize("x, y, z, expected", cases)
    def test(self, x, y, z, expected):
        assert self.table[x, y, z].eval(self.state, self.model) == expected

    union_cases = []

    for x in [set_const, set_expr, set_var]:
        for y in [set_const, set_expr, set_var]:
            for z in [set_const, set_expr, set_var]:
                union_cases.append((x, y, z, {0, 1, 2}))

    for value in [set_const, set_expr, set_var]:
        for zero in [zero_expr, zero_var, zero_resource_var, 0]:
            union_cases.append((value, value, zero, {0, 1, 2}))
            union_cases.append((value, zero, value, {0, 1, 2}))
            union_cases.append((value, zero, zero, {0, 1}))
            union_cases.append((zero, value, value, {0, 1, 2}))
            union_cases.append((zero, value, zero, {1, 2}))
            union_cases.append((zero, zero, value, {0, 1}))

    @pytest.mark.parametrize("x, y, z, expected", union_cases)
    def test_union(self, x, y, z, expected):
        assert self.table.union(x, y, z).eval(self.state, self.model) == expected

    intersection_cases = []

    for x in [set_const, set_expr, set_var]:
        for y in [set_const, set_expr, set_var]:
            for z in [set_const, set_expr, set_var]:
                intersection_cases.append((x, y, z, set()))

    for value in [set_const, set_expr, set_var]:
        for two in [two_expr, two_var, two_resource_var, 2]:
            for zero in [zero_expr, zero_var, zero_resource_var, 0]:
                intersection_cases.append((value, two, zero, {1}))

    for value in [set_const, set_expr, set_var]:
        for zero in [zero_expr, zero_var, zero_resource_var, 0]:
            intersection_cases.append((value, value, zero, set()))
            intersection_cases.append((value, zero, value, set()))
            intersection_cases.append((value, zero, zero, set()))
            intersection_cases.append((zero, value, value, set()))
            intersection_cases.append((zero, value, zero, set()))
            intersection_cases.append((zero, zero, value, set()))

    @pytest.mark.parametrize("x, y, z, expected", intersection_cases)
    def test_intersection(self, x, y, z, expected):
        assert self.table.intersection(x, y, z).eval(self.state, self.model) == expected

    symmetric_difference_cases = []

    for x in [set_const, set_expr, set_var]:
        for y in [set_const, set_expr, set_var]:
            for z in [set_const, set_expr, set_var]:
                symmetric_difference_cases.append((x, y, z, set()))

    for value in [set_const, set_expr, set_var]:
        for zero in [zero_expr, zero_var, zero_resource_var, 0]:
            symmetric_difference_cases.append((value, value, zero, {1, 2}))
            symmetric_difference_cases.append((value, zero, value, {1, 2}))
            symmetric_difference_cases.append((value, zero, zero, {0, 1}))
            symmetric_difference_cases.append((zero, value, value, {1, 2}))
            symmetric_difference_cases.append((zero, value, zero, {1, 2}))
            symmetric_difference_cases.append((zero, zero, value, {0, 1}))

    @pytest.mark.parametrize("x, y, z, expected", symmetric_difference_cases)
    def test_symmetric_difference(self, x, y, z, expected):
        assert (
            self.table.symmetric_difference(x, y, z).eval(self.state, self.model)
            == expected
        )

    error_cases = [
        0,
        (0, 1),
        (-1, 0, 0),
        (0, -1, 0),
        (0, 0, -1),
        (dp.IntExpr(1), 0, 0),
        (0, dp.IntExpr(1), 0),
        (0, 0, dp.IntExpr(1)),
        (3, 0, 0),
        (0, 3, 0),
        (0, 0, 3),
    ]

    @pytest.mark.parametrize("index", error_cases)
    def test_error(self, index):
        with pytest.raises(BaseException):
            self.table[index].eval(self.state, self.model)


class TestSetTable:
    model = dp.Model()
    obj = model.add_object_type(number=3)

    zero_expr = dp.ElementExpr(0)
    zero_var = model.add_element_var(object_type=obj, target=0)
    zero_resource_var = model.add_element_resource_var(
        object_type=obj, target=0, less_is_better=True
    )

    one_expr = dp.ElementExpr(1)
    one_var = model.add_element_var(object_type=obj, target=1)
    one_resource_var = model.add_element_resource_var(
        object_type=obj, target=1, less_is_better=True
    )

    set_const = model.create_set_const(object_type=obj, value=[0, 1])
    set_expr = dp.SetExpr(set_const)
    set_var = model.add_set_var(object_type=obj, target=[0, 1])

    table = model.add_set_table(
        {(0, 0, 0, 0): [0, 1], (1, 1, 1, 1): [1, 2]}, default=[0, 2], object_type=obj
    )
    state = model.target_state

    cases = (
        [
            ((zero, zero, zero, zero), {0, 1})
            for zero in [zero_expr, zero_var, zero_resource_var, 0]
        ]
        + [
            ((one, one, one, one), {1, 2})
            for one in [one_expr, one_var, one_resource_var, 1]
        ]
        + [((0, 1, 0, 1), {0, 2})]
    )

    @pytest.mark.parametrize("index, expected", cases)
    def test(self, index, expected):
        assert self.table[index].eval(self.state, self.model) == expected

    union_cases = [
        ((value, value, value, value), {0, 1, 2})
        for value in [set_const, set_expr, set_var]
    ]

    for value in [set_const, set_expr, set_var]:
        for zero in [zero_expr, zero_var, zero_resource_var, 0]:
            union_cases.append(((zero, zero, zero, value), {0, 1, 2}))

    @pytest.mark.parametrize("index, expected", union_cases)
    def test_union(self, index, expected):
        assert self.table.union(index).eval(self.state, self.model) == expected

    intersection_cases = [
        ((value, value, value, value), set())
        for value in [set_const, set_expr, set_var]
    ]

    for value in [set_const, set_expr, set_var]:
        for zero in [zero_expr, zero_var, zero_resource_var, 0]:
            intersection_cases.append(((zero, zero, zero, value), {0}))

    @pytest.mark.parametrize("index, expected", intersection_cases)
    def test_intersection(self, index, expected):
        assert self.table.intersection(index).eval(self.state, self.model) == expected

    symmetric_difference_cases = [
        ((value, value, value, value), {0, 2})
        for value in [set_const, set_expr, set_var]
    ]

    for value in [set_const, set_expr, set_var]:
        for zero in [zero_expr, zero_var, zero_resource_var, 0]:
            symmetric_difference_cases.append(((zero, zero, zero, value), {1, 2}))

    @pytest.mark.parametrize("index, expected", symmetric_difference_cases)
    def test_symmetric_difference(self, index, expected):
        assert (
            self.table.symmetric_difference(index).eval(self.state, self.model)
            == expected
        )

    error_cases = [
        (-1, 0, 0),
        (0, -1, 0),
        (0, 0, -1),
        (dp.IntExpr(1), 0, 0),
        (0, dp.IntExpr(1), 0),
        (0, 0, dp.IntExpr(1)),
    ]

    @pytest.mark.parametrize("index", error_cases)
    def test_error(self, index):
        with pytest.raises(BaseException):
            self.table[index].eval(self.state, self.model)


class TestBoolTable1D:
    model = dp.Model()
    obj = model.add_object_type(number=2)

    zero_expr = dp.ElementExpr(0)
    zero_var = model.add_element_var(object_type=obj, target=0)
    zero_resource_var = model.add_element_resource_var(
        object_type=obj, target=0, less_is_better=True
    )

    one_expr = dp.ElementExpr(1)
    one_var = model.add_element_var(object_type=obj, target=1)
    one_resource_var = model.add_element_resource_var(
        object_type=obj, target=1, less_is_better=True
    )

    table = model.add_bool_table([True, False])
    state = model.target_state

    cases = [
        (zero_expr, True),
        (zero_var, True),
        (zero_resource_var, True),
        (0, True),
        (one_expr, False),
        (one_var, False),
        (one_resource_var, False),
        (1, False),
    ]

    @pytest.mark.parametrize("x, expected", cases)
    def test(self, x, expected):
        assert self.table[x].eval(self.state, self.model) == expected

    error_cases = [-1, dp.IntExpr(0), 2]

    @pytest.mark.parametrize("index", error_cases)
    def test_error(self, index):
        with pytest.raises(BaseException):
            self.table[index].eval(self.state, self.model)


class TestBoolTable2D:
    model = dp.Model()
    obj = model.add_object_type(number=2)

    zero_expr = dp.ElementExpr(0)
    zero_var = model.add_element_var(object_type=obj, target=0)
    zero_resource_var = model.add_element_resource_var(
        object_type=obj, target=0, less_is_better=True
    )

    one_expr = dp.ElementExpr(1)
    one_var = model.add_element_var(object_type=obj, target=1)
    one_resource_var = model.add_element_resource_var(
        object_type=obj, target=1, less_is_better=True
    )

    table = model.add_bool_table([[True, False], [False, True]])
    state = model.target_state

    cases = [
        (zero, zero, True) for zero in [zero_expr, zero_var, zero_resource_var, 0]
    ] + [(one, one, True) for one in [one_expr, one_var, one_resource_var, 1]]

    for zero in [zero_expr, zero_var, zero_resource_var, 0]:
        for one in [one_expr, one_var, one_resource_var, 1]:
            cases.append((zero, one, False))
            cases.append((one, zero, False))

    @pytest.mark.parametrize("x, y, expected", cases)
    def test(self, x, y, expected):
        assert self.table[x, y].eval(self.state, self.model) == expected

    error_cases = [
        0,
        (0, -1),
        (-1, 0),
        (0, dp.IntExpr(1)),
        (dp.IntExpr(1), 0),
        (0, 2),
        (2, 0),
    ]

    @pytest.mark.parametrize("index", error_cases)
    def test_error(self, index):
        with pytest.raises(BaseException):
            self.table[index].eval(self.state, self.model)


class TestBoolTable3D:
    model = dp.Model()
    obj = model.add_object_type(number=3)

    zero_expr = dp.ElementExpr(0)
    zero_var = model.add_element_var(object_type=obj, target=0)
    zero_resource_var = model.add_element_resource_var(
        object_type=obj, target=0, less_is_better=True
    )

    one_expr = dp.ElementExpr(1)
    one_var = model.add_element_var(object_type=obj, target=1)
    one_resource_var = model.add_element_resource_var(
        object_type=obj, target=1, less_is_better=True
    )

    two_expr = dp.ElementExpr(2)
    two_var = model.add_element_var(object_type=obj, target=2)
    two_resource_var = model.add_element_resource_var(
        object_type=obj, target=2, less_is_better=True
    )

    table = model.add_bool_table(
        [
            [[True, False, True], [True, False, True], [True, True, False]],
            [[False, True, True], [False, True, True], [True, False, True]],
            [[True, True, False], [False, True, True], [False, True, True]],
        ]
    )
    state = model.target_state

    cases = (
        [(zero, zero, zero, 1) for zero in [zero_expr, zero_var, zero_resource_var, 0]]
        + [(one, one, one, 1) for one in [one_expr, one_var, one_resource_var, 1]]
        + [(two, two, two, 1) for two in [two_expr, two_var, two_resource_var, 2]]
    )

    for zero in [zero_expr, zero_var, zero_resource_var, 0]:
        for one in [one_expr, one_var, one_resource_var, 1]:
            for two in [two_expr, two_var, two_resource_var, 2]:
                cases.append((zero, one, two, True))
                cases.append((zero, two, one, True))
                cases.append((one, zero, two, True))
                cases.append((one, two, zero, True))
                cases.append((two, zero, one, True))
                cases.append((two, one, zero, False))

    for zero in [zero_expr, zero_var, zero_resource_var, 0]:
        for one in [one_expr, one_var, one_resource_var, 1]:
            cases.append((zero, zero, one, False))
            cases.append((zero, one, zero, True))
            cases.append((zero, one, one, False))
            cases.append((one, zero, zero, False))
            cases.append((one, zero, one, True))
            cases.append((one, one, zero, False))

    for zero in [zero_expr, zero_var, zero_resource_var, 0]:
        for two in [two_expr, two_var, two_resource_var, 2]:
            cases.append((zero, zero, two, True))
            cases.append((zero, two, zero, True))
            cases.append((zero, two, two, False))
            cases.append((two, zero, zero, True))
            cases.append((two, zero, two, False))
            cases.append((two, two, zero, False))

    for one in [one_expr, one_var, one_resource_var, 1]:
        for two in [two_expr, two_var, two_resource_var, 2]:
            cases.append((one, one, two, True))
            cases.append((one, two, one, False))
            cases.append((one, two, two, True))
            cases.append((two, one, one, True))
            cases.append((two, one, two, True))
            cases.append((two, two, one, True))

    @pytest.mark.parametrize("x, y, z, expected", cases)
    def test(self, x, y, z, expected):
        assert self.table[x, y, z].eval(self.state, self.model) == expected

    error_cases = [
        0,
        (0, 1),
        (-1, 0, 0),
        (0, -1, 0),
        (0, 0, -1),
        (dp.IntExpr(1), 0, 0),
        (0, dp.IntExpr(1), 0),
        (0, 0, dp.IntExpr(1)),
        (3, 0, 0),
        (0, 3, 0),
        (0, 0, 3),
    ]

    @pytest.mark.parametrize("index", error_cases)
    def test_error(self, index):
        with pytest.raises(BaseException):
            self.table[index].eval(self.state, self.model)


class TestBoolTable:
    model = dp.Model()
    obj = model.add_object_type(number=3)

    zero_expr = dp.ElementExpr(0)
    zero_var = model.add_element_var(object_type=obj, target=0)
    zero_resource_var = model.add_element_resource_var(
        object_type=obj, target=0, less_is_better=True
    )

    one_expr = dp.ElementExpr(1)
    one_var = model.add_element_var(object_type=obj, target=1)
    one_resource_var = model.add_element_resource_var(
        object_type=obj, target=1, less_is_better=True
    )

    table = model.add_bool_table(
        {(0, 0, 0, 0): True, (1, 1, 1, 1): False}, default=True
    )
    state = model.target_state

    cases = (
        [
            ((zero, zero, zero, zero), True)
            for zero in [zero_expr, zero_var, zero_resource_var, 0]
        ]
        + [
            ((one, one, one, one), False)
            for one in [one_expr, one_var, one_resource_var, 1]
        ]
        + [((0, 1, 0, 1), True)]
    )

    @pytest.mark.parametrize("index, expected", cases)
    def test(self, index, expected):
        assert self.table[index].eval(self.state, self.model) == expected

    error_cases = [
        (-1, 0, 0),
        (0, -1, 0),
        (0, 0, -1),
        (dp.IntExpr(1), 0, 0),
        (0, dp.IntExpr(1), 0),
        (0, 0, dp.IntExpr(1)),
    ]

    @pytest.mark.parametrize("index", error_cases)
    def test_error(self, index):
        with pytest.raises(BaseException):
            self.table[index].eval(self.state, self.model)


class TestIntTable1D:
    model = dp.Model()
    obj = model.add_object_type(number=2)

    zero_expr = dp.ElementExpr(0)
    zero_var = model.add_element_var(object_type=obj, target=0)
    zero_resource_var = model.add_element_resource_var(
        object_type=obj, target=0, less_is_better=True
    )

    one_expr = dp.ElementExpr(1)
    one_var = model.add_element_var(object_type=obj, target=1)
    one_resource_var = model.add_element_resource_var(
        object_type=obj, target=1, less_is_better=True
    )

    set_const = model.create_set_const(object_type=obj, value=[0, 1])
    set_expr = dp.SetExpr(set_const)
    set_var = model.add_set_var(object_type=obj, target=[0, 1])

    table = model.add_int_table([3, 2])
    state = model.target_state

    cases = [
        (zero_expr, 3),
        (zero_var, 3),
        (zero_resource_var, 3),
        (0, 3),
        (one_expr, 2),
        (one_var, 2),
        (one_resource_var, 2),
        (1, 2),
        (set_const, 5),
        (set_expr, 5),
        (set_var, 5),
    ]

    @pytest.mark.parametrize("x, expected", cases)
    def test(self, x, expected):
        assert self.table[x].eval(self.state, self.model) == expected

    product_cases = [(set_const, 6), (set_expr, 6), (set_var, 6)]

    @pytest.mark.parametrize("x, expected", product_cases)
    def test_product(self, x, expected):
        assert self.table.product(x).eval(self.state, self.model) == expected

    max_cases = [(set_const, 3), (set_expr, 3), (set_var, 3)]

    @pytest.mark.parametrize("x, expected", max_cases)
    def test_max(self, x, expected):
        assert self.table.max(x).eval(self.state, self.model) == expected

    min_cases = [(set_const, 2), (set_expr, 2), (set_var, 2)]

    @pytest.mark.parametrize("x, expected", min_cases)
    def test_min(self, x, expected):
        assert self.table.min(x).eval(self.state, self.model) == expected

    error_cases = [-1, dp.IntExpr(0), 2]

    @pytest.mark.parametrize("index", error_cases)
    def test_error(self, index):
        with pytest.raises(BaseException):
            self.table[index].eval(self.state, self.model)


class TestIntTable2D:
    model = dp.Model()
    obj = model.add_object_type(number=2)

    zero_expr = dp.ElementExpr(0)
    zero_var = model.add_element_var(object_type=obj, target=0)
    zero_resource_var = model.add_element_resource_var(
        object_type=obj, target=0, less_is_better=True
    )

    one_expr = dp.ElementExpr(1)
    one_var = model.add_element_var(object_type=obj, target=1)
    one_resource_var = model.add_element_resource_var(
        object_type=obj, target=1, less_is_better=True
    )

    set_const = model.create_set_const(object_type=obj, value=[0, 1])
    set_expr = dp.SetExpr(set_const)
    set_var = model.add_set_var(object_type=obj, target=[0, 1])

    table = model.add_int_table([[3, 2], [2, 3]])
    state = model.target_state

    cases = [
        (zero, zero, 3) for zero in [zero_expr, zero_var, zero_resource_var, 0]
    ] + [(one, one, 3) for one in [one_expr, one_var, one_resource_var, 1]]

    for x in [set_const, set_expr, set_var]:
        for y in [set_const, set_expr, set_var]:
            cases.append((x, y, 10))

    for zero in [zero_expr, zero_var, zero_resource_var, 0]:
        for one in [one_expr, one_var, one_resource_var, 1]:
            cases.append((zero, one, 2))
            cases.append((one, zero, 2))

    for value in [set_const, set_expr, set_var]:
        for zero in [zero_expr, zero_var, zero_resource_var, 0]:
            cases.append((value, zero, 5))
            cases.append((zero, value, 5))

    @pytest.mark.parametrize("x, y, expected", cases)
    def test(self, x, y, expected):
        assert self.table[x, y].eval(self.state, self.model) == expected

    product_cases = []

    for x in [set_const, set_expr, set_var]:
        for y in [set_const, set_expr, set_var]:
            product_cases.append((x, y, 36))

    for value in [set_const, set_expr, set_var]:
        for zero in [zero_expr, zero_var, zero_resource_var, 0]:
            product_cases.append((value, zero, 6))
            product_cases.append((zero, value, 6))

    @pytest.mark.parametrize("x, y, expected", product_cases)
    def test_product(self, x, y, expected):
        assert self.table.product(x, y).eval(self.state, self.model) == expected

    max_cases = []

    for x in [set_const, set_expr, set_var]:
        for y in [set_const, set_expr, set_var]:
            max_cases.append((x, y, 3))

    for value in [set_const, set_expr, set_var]:
        for zero in [zero_expr, zero_var, zero_resource_var, 0]:
            max_cases.append((value, zero, 3))
            max_cases.append((zero, value, 3))

    @pytest.mark.parametrize("x, y, expected", max_cases)
    def test_max(self, x, y, expected):
        assert self.table.max(x, y).eval(self.state, self.model) == expected

    min_cases = [(value, value, 2) for value in [set_const, set_expr, set_var]]

    for x in [set_const, set_expr, set_var]:
        for y in [set_const, set_expr, set_var]:
            min_cases.append((x, y, 2))

    for value in [set_const, set_expr, set_var]:
        for zero in [zero_expr, zero_var, zero_resource_var, 0]:
            min_cases.append((value, zero, 2))
            min_cases.append((zero, value, 2))

    @pytest.mark.parametrize("x, y, expected", min_cases)
    def test_min(self, x, y, expected):
        assert self.table.min(x, y).eval(self.state, self.model) == expected

    error_cases = [
        0,
        (0, -1),
        (-1, 0),
        (0, dp.IntExpr(1)),
        (dp.IntExpr(1), 0),
        (0, 2),
        (2, 0),
    ]

    @pytest.mark.parametrize("index", error_cases)
    def test_error(self, index):
        with pytest.raises(BaseException):
            self.table[index].eval(self.state, self.model)


class TestIntTable3D:
    model = dp.Model()
    obj = model.add_object_type(number=3)

    zero_expr = dp.ElementExpr(0)
    zero_var = model.add_element_var(object_type=obj, target=0)
    zero_resource_var = model.add_element_resource_var(
        object_type=obj, target=0, less_is_better=True
    )

    one_expr = dp.ElementExpr(1)
    one_var = model.add_element_var(object_type=obj, target=1)
    one_resource_var = model.add_element_resource_var(
        object_type=obj, target=1, less_is_better=True
    )

    set_const = model.create_set_const(object_type=obj, value=[0, 1])
    set_expr = dp.SetExpr(set_const)
    set_var = model.add_set_var(object_type=obj, target=[0, 1])

    two_expr = dp.ElementExpr(2)
    two_var = model.add_element_var(object_type=obj, target=2)
    two_resource_var = model.add_element_resource_var(
        object_type=obj, target=2, less_is_better=True
    )

    table = model.add_int_table(
        [
            [[3, 2, 4], [4, 2, 3], [3, 4, 2]],
            [[2, 4, 3], [2, 3, 4], [3, 2, 4]],
            [[3, 4, 2], [2, 3, 4], [2, 4, 3]],
        ]
    )
    state = model.target_state

    cases = (
        [(zero, zero, zero, 3) for zero in [zero_expr, zero_var, zero_resource_var, 0]]
        + [(one, one, one, 3) for one in [one_expr, one_var, one_resource_var, 1]]
        + [(two, two, two, 3) for two in [two_expr, two_var, two_resource_var, 2]]
    )

    for x in [set_const, set_expr, set_var]:
        for y in [set_const, set_expr, set_var]:
            for z in [set_const, set_expr, set_var]:
                cases.append((x, y, z, 22))

    for zero in [zero_expr, zero_var, zero_resource_var, 0]:
        for one in [one_expr, one_var, one_resource_var, 1]:
            for two in [two_expr, two_var, two_resource_var, 2]:
                cases.append((zero, one, two, 3))
                cases.append((zero, two, one, 4))
                cases.append((one, zero, two, 3))
                cases.append((one, two, zero, 3))
                cases.append((two, zero, one, 4))
                cases.append((two, one, zero, 2))

    for zero in [zero_expr, zero_var, zero_resource_var, 0]:
        for one in [one_expr, one_var, one_resource_var, 1]:
            cases.append((zero, zero, one, 2))
            cases.append((zero, one, zero, 4))
            cases.append((zero, one, one, 2))
            cases.append((one, zero, zero, 2))
            cases.append((one, zero, one, 4))
            cases.append((one, one, zero, 2))

    for zero in [zero_expr, zero_var, zero_resource_var, 0]:
        for two in [two_expr, two_var, two_resource_var, 2]:
            cases.append((zero, zero, two, 4))
            cases.append((zero, two, zero, 3))
            cases.append((zero, two, two, 2))
            cases.append((two, zero, zero, 3))
            cases.append((two, zero, two, 2))
            cases.append((two, two, zero, 2))

    for one in [one_expr, one_var, one_resource_var, 1]:
        for two in [two_expr, two_var, two_resource_var, 2]:
            cases.append((one, one, two, 4))
            cases.append((one, two, one, 2))
            cases.append((one, two, two, 4))
            cases.append((two, one, one, 3))
            cases.append((two, one, two, 4))
            cases.append((two, two, one, 4))

    for value in [set_const, set_expr, set_var]:
        for zero in [zero_expr, zero_var, zero_resource_var, 0]:
            cases.append((value, value, zero, 11))
            cases.append((value, zero, value, 11))
            cases.append((value, zero, zero, 5))
            cases.append((zero, value, value, 11))
            cases.append((zero, value, zero, 7))
            cases.append((zero, zero, value, 5))

    @pytest.mark.parametrize("x, y, z, expected", cases)
    def test(self, x, y, z, expected):
        assert self.table[x, y, z].eval(self.state, self.model) == expected

    product_cases = []

    for x in [set_const, set_expr, set_var]:
        for y in [set_const, set_expr, set_var]:
            for z in [set_const, set_expr, set_var]:
                product_cases.append((x, y, z, 2304))

    for value in [set_const, set_expr, set_var]:
        for zero in [zero_expr, zero_var, zero_resource_var, 0]:
            product_cases.append((value, value, zero, 48))
            product_cases.append((value, zero, value, 48))
            product_cases.append((value, zero, zero, 6))
            product_cases.append((zero, value, value, 48))
            product_cases.append((zero, value, zero, 12))
            product_cases.append((zero, zero, value, 6))

    @pytest.mark.parametrize("x, y, z, expected", product_cases)
    def test_product(self, x, y, z, expected):
        assert self.table.product(x, y, z).eval(self.state, self.model) == expected

    max_cases = []

    for x in [set_const, set_expr, set_var]:
        for y in [set_const, set_expr, set_var]:
            for z in [set_const, set_expr, set_var]:
                max_cases.append((x, y, z, 4))

    for value in [set_const, set_expr, set_var]:
        for zero in [zero_expr, zero_var, zero_resource_var, 0]:
            max_cases.append((value, value, zero, 4))
            max_cases.append((value, zero, value, 4))
            max_cases.append((value, zero, zero, 3))
            max_cases.append((zero, value, value, 4))
            max_cases.append((zero, value, zero, 4))
            max_cases.append((zero, zero, value, 3))

    @pytest.mark.parametrize("x, y, z, expected", max_cases)
    def test_max(self, x, y, z, expected):
        assert self.table.max(x, y, z).eval(self.state, self.model) == expected

    min_cases = []

    for x in [set_const, set_expr, set_var]:
        for y in [set_const, set_expr, set_var]:
            for z in [set_const, set_expr, set_var]:
                min_cases.append((x, y, z, 2))

    for value in [set_const, set_expr, set_var]:
        for zero in [zero_expr, zero_var, zero_resource_var, 0]:
            min_cases.append((value, value, zero, 2))
            min_cases.append((value, zero, value, 2))
            min_cases.append((value, zero, zero, 2))
            min_cases.append((zero, value, value, 2))
            min_cases.append((zero, value, zero, 3))
            min_cases.append((zero, zero, value, 2))

    @pytest.mark.parametrize("x, y, z, expected", min_cases)
    def test_min(self, x, y, z, expected):
        assert self.table.min(x, y, z).eval(self.state, self.model) == expected

    error_cases = [
        0,
        (0, 1),
        (-1, 0, 0),
        (0, -1, 0),
        (0, 0, -1),
        (dp.IntExpr(1), 0, 0),
        (0, dp.IntExpr(1), 0),
        (0, 0, dp.IntExpr(1)),
        (3, 0, 0),
        (0, 3, 0),
        (0, 0, 3),
    ]

    @pytest.mark.parametrize("index", error_cases)
    def test_error(self, index):
        with pytest.raises(BaseException):
            self.table[index].eval(self.state, self.model)


class TestIntTable:
    model = dp.Model()
    obj = model.add_object_type(number=3)

    zero_expr = dp.ElementExpr(0)
    zero_var = model.add_element_var(object_type=obj, target=0)
    zero_resource_var = model.add_element_resource_var(
        object_type=obj, target=0, less_is_better=True
    )

    one_expr = dp.ElementExpr(1)
    one_var = model.add_element_var(object_type=obj, target=1)
    one_resource_var = model.add_element_resource_var(
        object_type=obj, target=1, less_is_better=True
    )

    set_const = model.create_set_const(object_type=obj, value=[0, 1])
    set_expr = dp.SetExpr(set_const)
    set_var = model.add_set_var(object_type=obj, target=[0, 1])

    table = model.add_int_table({(0, 0, 0, 0): 3, (1, 1, 1, 1): 2}, default=4)
    state = model.target_state

    cases = (
        [
            ((zero, zero, zero, zero), 3)
            for zero in [zero_expr, zero_var, zero_resource_var, 0]
        ]
        + [
            ((one, one, one, one), 2)
            for one in [one_expr, one_var, one_resource_var, 1]
        ]
        + [((0, 1, 0, 1), 4)]
        + [
            ((value, value, value, value), 61)
            for value in [set_const, set_expr, set_var]
        ]
    )

    for value in [set_const, set_expr, set_var]:
        for zero in [zero_expr, zero_var, zero_resource_var, 0]:
            cases.append(((zero, zero, zero, value), 7))

    @pytest.mark.parametrize("index, expected", cases)
    def test(self, index, expected):
        assert self.table[index].eval(self.state, self.model) == expected

    product_cases = [
        ((value, value, value, value), 1610612736)
        for value in [set_const, set_expr, set_var]
    ]

    for value in [set_const, set_expr, set_var]:
        for zero in [zero_expr, zero_var, zero_resource_var, 0]:
            product_cases.append(((zero, zero, zero, value), 12))

    @pytest.mark.parametrize("index, expected", product_cases)
    def test_product(self, index, expected):
        assert self.table.product(index).eval(self.state, self.model) == expected

    max_cases = [
        ((value, value, value, value), 4) for value in [set_const, set_expr, set_var]
    ]

    for value in [set_const, set_expr, set_var]:
        for zero in [zero_expr, zero_var, zero_resource_var, 0]:
            max_cases.append(((zero, zero, zero, value), 4))

    @pytest.mark.parametrize("index, expected", max_cases)
    def test_max(self, index, expected):
        assert self.table.max(index).eval(self.state, self.model) == expected

    min_cases = [
        ((value, value, value, value), 2) for value in [set_const, set_expr, set_var]
    ]

    for value in [set_const, set_expr, set_var]:
        for zero in [zero_expr, zero_var, zero_resource_var, 0]:
            min_cases.append(((zero, zero, zero, value), 3))

    @pytest.mark.parametrize("index, expected", min_cases)
    def test_min(self, index, expected):
        assert self.table.min(index).eval(self.state, self.model) == expected

    error_cases = [
        (-1, 0, 0),
        (0, -1, 0),
        (0, 0, -1),
        (dp.IntExpr(1), 0, 0),
        (0, dp.IntExpr(1), 0),
        (0, 0, dp.IntExpr(1)),
    ]

    @pytest.mark.parametrize("index", error_cases)
    def test_error(self, index):
        with pytest.raises(BaseException):
            self.table[index].eval(self.state, self.model)


class TestFloatTable1D:
    model = dp.Model()
    obj = model.add_object_type(number=2)

    zero_expr = dp.ElementExpr(0)
    zero_var = model.add_element_var(object_type=obj, target=0)
    zero_resource_var = model.add_element_resource_var(
        object_type=obj, target=0, less_is_better=True
    )

    one_expr = dp.ElementExpr(1)
    one_var = model.add_element_var(object_type=obj, target=1)
    one_resource_var = model.add_element_resource_var(
        object_type=obj, target=1, less_is_better=True
    )

    set_const = model.create_set_const(object_type=obj, value=[0, 1])
    set_expr = dp.SetExpr(set_const)
    set_var = model.add_set_var(object_type=obj, target=[0, 1])

    table = model.add_float_table([0.3, 0.2])
    state = model.target_state

    cases = [
        (zero_expr, pytest.approx(0.3)),
        (zero_var, pytest.approx(0.3)),
        (zero_resource_var, pytest.approx(0.3)),
        (0, pytest.approx(0.3)),
        (one_expr, pytest.approx(0.2)),
        (one_var, pytest.approx(0.2)),
        (one_resource_var, pytest.approx(0.2)),
        (1, pytest.approx(0.2)),
        (set_const, pytest.approx(0.5)),
        (set_expr, pytest.approx(0.5)),
        (set_var, pytest.approx(0.5)),
    ]

    @pytest.mark.parametrize("x, expected", cases)
    def test(self, x, expected):
        assert self.table[x].eval(self.state, self.model) == expected

    product_cases = [
        (set_const, pytest.approx(0.06)),
        (set_expr, pytest.approx(0.06)),
        (set_var, pytest.approx(0.06)),
    ]

    @pytest.mark.parametrize("x, expected", product_cases)
    def test_product(self, x, expected):
        assert self.table.product(x).eval(self.state, self.model) == expected

    max_cases = [
        (set_const, pytest.approx(0.3)),
        (set_expr, pytest.approx(0.3)),
        (set_var, pytest.approx(0.3)),
    ]

    @pytest.mark.parametrize("x, expected", max_cases)
    def test_max(self, x, expected):
        assert self.table.max(x).eval(self.state, self.model) == expected

    min_cases = [
        (set_const, pytest.approx(0.2)),
        (set_expr, pytest.approx(0.2)),
        (set_var, pytest.approx(0.2)),
    ]

    @pytest.mark.parametrize("x, expected", min_cases)
    def test_min(self, x, expected):
        assert self.table.min(x).eval(self.state, self.model) == expected

    error_cases = [-1, dp.FloatExpr(0), 2]

    @pytest.mark.parametrize("index", error_cases)
    def test_error(self, index):
        with pytest.raises(BaseException):
            self.table[index].eval(self.state, self.model)


class TestFloatTable2D:
    model = dp.Model()
    obj = model.add_object_type(number=2)

    zero_expr = dp.ElementExpr(0)
    zero_var = model.add_element_var(object_type=obj, target=0)
    zero_resource_var = model.add_element_resource_var(
        object_type=obj, target=0, less_is_better=True
    )

    one_expr = dp.ElementExpr(1)
    one_var = model.add_element_var(object_type=obj, target=1)
    one_resource_var = model.add_element_resource_var(
        object_type=obj, target=1, less_is_better=True
    )

    set_const = model.create_set_const(object_type=obj, value=[0, 1])
    set_expr = dp.SetExpr(set_const)
    set_var = model.add_set_var(object_type=obj, target=[0, 1])

    table = model.add_float_table([[0.3, 0.2], [0.2, 0.3]])
    state = model.target_state

    cases = [
        (zero, zero, pytest.approx(0.3))
        for zero in [zero_expr, zero_var, zero_resource_var, 0]
    ] + [
        (one, one, pytest.approx(0.3))
        for one in [one_expr, one_var, one_resource_var, 1]
    ]

    for x in [set_const, set_expr, set_var]:
        for y in [set_const, set_expr, set_var]:
            cases.append((x, y, pytest.approx(1.0)))

    for zero in [zero_expr, zero_var, zero_resource_var, 0]:
        for one in [one_expr, one_var, one_resource_var, 1]:
            cases.append((zero, one, pytest.approx(0.2)))
            cases.append((one, zero, pytest.approx(0.2)))

    for value in [set_const, set_expr, set_var]:
        for zero in [zero_expr, zero_var, zero_resource_var, 0]:
            cases.append((value, zero, pytest.approx(0.5)))
            cases.append((zero, value, pytest.approx(0.5)))

    @pytest.mark.parametrize("x, y, expected", cases)
    def test(self, x, y, expected):
        assert self.table[x, y].eval(self.state, self.model) == expected

    product_cases = []

    for x in [set_const, set_expr, set_var]:
        for y in [set_const, set_expr, set_var]:
            product_cases.append((x, y, pytest.approx(0.0036)))

    for value in [set_const, set_expr, set_var]:
        for zero in [zero_expr, zero_var, zero_resource_var, 0]:
            product_cases.append((value, zero, pytest.approx(0.06)))
            product_cases.append((zero, value, pytest.approx(0.06)))

    @pytest.mark.parametrize("x, y, expected", product_cases)
    def test_product(self, x, y, expected):
        assert self.table.product(x, y).eval(self.state, self.model) == expected

    max_cases = []

    for x in [set_const, set_expr, set_var]:
        for y in [set_const, set_expr, set_var]:
            max_cases.append((x, y, pytest.approx(0.3)))

    for value in [set_const, set_expr, set_var]:
        for zero in [zero_expr, zero_var, zero_resource_var, 0]:
            max_cases.append((value, zero, pytest.approx(0.3)))
            max_cases.append((zero, value, pytest.approx(0.3)))

    @pytest.mark.parametrize("x, y, expected", max_cases)
    def test_max(self, x, y, expected):
        assert self.table.max(x, y).eval(self.state, self.model) == expected

    min_cases = []

    for x in [set_const, set_expr, set_var]:
        for y in [set_const, set_expr, set_var]:
            min_cases.append((x, y, pytest.approx(0.2)))

    for value in [set_const, set_expr, set_var]:
        for zero in [zero_expr, zero_var, zero_resource_var, 0]:
            min_cases.append((value, zero, pytest.approx(0.2)))
            min_cases.append((zero, value, pytest.approx(0.2)))

    @pytest.mark.parametrize("x, y, expected", min_cases)
    def test_min(self, x, y, expected):
        assert self.table.min(x, y).eval(self.state, self.model) == expected

    error_cases = [
        0,
        (0, -1),
        (-1, 0),
        (0, dp.FloatExpr(1)),
        (dp.FloatExpr(1), 0),
        (0, 2),
        (2, 0),
    ]

    @pytest.mark.parametrize("index", error_cases)
    def test_error(self, index):
        with pytest.raises(BaseException):
            self.table[index].eval(self.state, self.model)


class TestFloatTable3D:
    model = dp.Model()
    obj = model.add_object_type(number=3)

    zero_expr = dp.ElementExpr(0)
    zero_var = model.add_element_var(object_type=obj, target=0)
    zero_resource_var = model.add_element_resource_var(
        object_type=obj, target=0, less_is_better=True
    )

    one_expr = dp.ElementExpr(1)
    one_var = model.add_element_var(object_type=obj, target=1)
    one_resource_var = model.add_element_resource_var(
        object_type=obj, target=1, less_is_better=True
    )

    set_const = model.create_set_const(object_type=obj, value=[0, 1])
    set_expr = dp.SetExpr(set_const)
    set_var = model.add_set_var(object_type=obj, target=[0, 1])

    two_expr = dp.ElementExpr(2)
    two_var = model.add_element_var(object_type=obj, target=2)
    two_resource_var = model.add_element_resource_var(
        object_type=obj, target=2, less_is_better=True
    )

    table = model.add_float_table(
        [
            [[0.3, 0.2, 0.4], [0.4, 0.2, 0.3], [0.3, 0.4, 0.2]],
            [[0.2, 0.4, 0.3], [0.2, 0.3, 0.4], [0.3, 0.2, 0.4]],
            [[0.3, 0.4, 0.2], [0.2, 0.3, 0.4], [0.2, 0.4, 0.3]],
        ]
    )
    state = model.target_state

    cases = (
        [
            (zero, zero, zero, pytest.approx(0.3))
            for zero in [zero_expr, zero_var, zero_resource_var, 0]
        ]
        + [
            (one, one, one, pytest.approx(0.3))
            for one in [one_expr, one_var, one_resource_var, 1]
        ]
        + [
            (two, two, two, pytest.approx(0.3))
            for two in [two_expr, two_var, two_resource_var, 2]
        ]
    )

    for x in [set_const, set_expr, set_var]:
        for y in [set_const, set_expr, set_var]:
            for z in [set_const, set_expr, set_var]:
                cases.append((x, y, z, pytest.approx(2.2)))

    for zero in [zero_expr, zero_var, zero_resource_var, 0]:
        for one in [one_expr, one_var, one_resource_var, 1]:
            for two in [two_expr, two_var, two_resource_var, 2]:
                cases.append((zero, one, two, pytest.approx(0.3)))
                cases.append((zero, two, one, pytest.approx(0.4)))
                cases.append((one, zero, two, pytest.approx(0.3)))
                cases.append((one, two, zero, pytest.approx(0.3)))
                cases.append((two, zero, one, pytest.approx(0.4)))
                cases.append((two, one, zero, pytest.approx(0.2)))

    for zero in [zero_expr, zero_var, zero_resource_var, 0]:
        for one in [one_expr, one_var, one_resource_var, 1]:
            cases.append((zero, zero, one, pytest.approx(0.2)))
            cases.append((zero, one, zero, pytest.approx(0.4)))
            cases.append((zero, one, one, pytest.approx(0.2)))
            cases.append((one, zero, zero, pytest.approx(0.2)))
            cases.append((one, zero, one, pytest.approx(0.4)))
            cases.append((one, one, zero, pytest.approx(0.2)))

    for zero in [zero_expr, zero_var, zero_resource_var, 0]:
        for two in [two_expr, two_var, two_resource_var, 2]:
            cases.append((zero, zero, two, pytest.approx(0.4)))
            cases.append((zero, two, zero, pytest.approx(0.3)))
            cases.append((zero, two, two, pytest.approx(0.2)))
            cases.append((two, zero, zero, pytest.approx(0.3)))
            cases.append((two, zero, two, pytest.approx(0.2)))
            cases.append((two, two, zero, pytest.approx(0.2)))

    for one in [one_expr, one_var, one_resource_var, 1]:
        for two in [two_expr, two_var, two_resource_var, 2]:
            cases.append((one, one, two, pytest.approx(0.4)))
            cases.append((one, two, one, pytest.approx(0.2)))
            cases.append((one, two, two, pytest.approx(0.4)))
            cases.append((two, one, one, pytest.approx(0.3)))
            cases.append((two, one, two, pytest.approx(0.4)))
            cases.append((two, two, one, pytest.approx(0.4)))

    for value in [set_const, set_expr, set_var]:
        for zero in [zero_expr, zero_var, zero_resource_var, 0]:
            cases.append((value, value, zero, pytest.approx(1.1)))
            cases.append((value, zero, value, pytest.approx(1.1)))
            cases.append((value, zero, zero, pytest.approx(0.5)))
            cases.append((zero, value, value, pytest.approx(1.1)))
            cases.append((zero, value, zero, pytest.approx(0.7)))
            cases.append((zero, zero, value, pytest.approx(0.5)))

    @pytest.mark.parametrize("x, y, z, expected", cases)
    def test(self, x, y, z, expected):
        assert self.table[x, y, z].eval(self.state, self.model) == expected

    product_cases = []

    for x in [set_const, set_expr, set_var]:
        for y in [set_const, set_expr, set_var]:
            for z in [set_const, set_expr, set_var]:
                product_cases.append((x, y, z, pytest.approx(2.304e-5)))

    for value in [set_const, set_expr, set_var]:
        for zero in [zero_expr, zero_var, zero_resource_var, 0]:
            product_cases.append((value, value, zero, pytest.approx(0.0048)))
            product_cases.append((value, zero, value, pytest.approx(0.0048)))
            product_cases.append((value, zero, zero, pytest.approx(0.06)))
            product_cases.append((zero, value, value, pytest.approx(0.0048)))
            product_cases.append((zero, value, zero, pytest.approx(0.12)))
            product_cases.append((zero, zero, value, pytest.approx(0.06)))

    @pytest.mark.parametrize("x, y, z, expected", product_cases)
    def test_product(self, x, y, z, expected):
        assert self.table.product(x, y, z).eval(self.state, self.model) == expected

    max_cases = []

    for x in [set_const, set_expr, set_var]:
        for y in [set_const, set_expr, set_var]:
            for z in [set_const, set_expr, set_var]:
                max_cases.append((x, y, z, pytest.approx(0.4)))

    for value in [set_const, set_expr, set_var]:
        for zero in [zero_expr, zero_var, zero_resource_var, 0]:
            max_cases.append((value, value, zero, pytest.approx(0.4)))
            max_cases.append((value, zero, value, pytest.approx(0.4)))
            max_cases.append((value, zero, zero, pytest.approx(0.3)))
            max_cases.append((zero, value, value, pytest.approx(0.4)))
            max_cases.append((zero, value, zero, pytest.approx(0.4)))
            max_cases.append((zero, zero, value, pytest.approx(0.3)))

    @pytest.mark.parametrize("x, y, z, expected", max_cases)
    def test_max(self, x, y, z, expected):
        assert self.table.max(x, y, z).eval(self.state, self.model) == expected

    min_cases = []

    for x in [set_const, set_expr, set_var]:
        for y in [set_const, set_expr, set_var]:
            for z in [set_const, set_expr, set_var]:
                min_cases.append((x, y, z, pytest.approx(0.2)))

    for value in [set_const, set_expr, set_var]:
        for zero in [zero_expr, zero_var, zero_resource_var, 0]:
            min_cases.append((value, value, zero, pytest.approx(0.2)))
            min_cases.append((value, zero, value, pytest.approx(0.2)))
            min_cases.append((value, zero, zero, pytest.approx(0.2)))
            min_cases.append((zero, value, value, pytest.approx(0.2)))
            min_cases.append((zero, value, zero, pytest.approx(0.3)))
            min_cases.append((zero, zero, value, pytest.approx(0.2)))

    @pytest.mark.parametrize("x, y, z, expected", min_cases)
    def test_min(self, x, y, z, expected):
        assert self.table.min(x, y, z).eval(self.state, self.model) == expected

    error_cases = [
        0,
        (0, 1),
        (-1, 0, 0),
        (0, -1, 0),
        (0, 0, -1),
        (dp.FloatExpr(1), 0, 0),
        (0, dp.FloatExpr(1), 0),
        (0, 0, dp.FloatExpr(1)),
        (3, 0, 0),
        (0, 3, 0),
        (0, 0, 3),
    ]

    @pytest.mark.parametrize("index", error_cases)
    def test_error(self, index):
        with pytest.raises(BaseException):
            self.table[index].eval(self.state, self.model)


class TestFloatTable:
    model = dp.Model()
    obj = model.add_object_type(number=3)

    zero_expr = dp.ElementExpr(0)
    zero_var = model.add_element_var(object_type=obj, target=0)
    zero_resource_var = model.add_element_resource_var(
        object_type=obj, target=0, less_is_better=True
    )

    one_expr = dp.ElementExpr(1)
    one_var = model.add_element_var(object_type=obj, target=1)
    one_resource_var = model.add_element_resource_var(
        object_type=obj, target=1, less_is_better=True
    )

    set_const = model.create_set_const(object_type=obj, value=[0, 1])
    set_expr = dp.SetExpr(set_const)
    set_var = model.add_set_var(object_type=obj, target=[0, 1])

    table = model.add_float_table({(0, 0, 0, 0): 0.3, (1, 1, 1, 1): 0.2}, default=0.4)
    state = model.target_state

    cases = (
        [
            ((zero, zero, zero, zero), pytest.approx(0.3))
            for zero in [zero_expr, zero_var, zero_resource_var, 0]
        ]
        + [
            ((one, one, one, one), pytest.approx(0.2))
            for one in [one_expr, one_var, one_resource_var, 1]
        ]
        + [((0, 1, 0, 1), pytest.approx(0.4))]
        + [
            ((value, value, value, value), pytest.approx(6.1))
            for value in [set_const, set_expr, set_var]
        ]
    )

    for value in [set_const, set_expr, set_var]:
        for zero in [zero_expr, zero_var, zero_resource_var, 0]:
            cases.append(((zero, zero, zero, value), pytest.approx(0.7)))

    @pytest.mark.parametrize("index, expected", cases)
    def test(self, index, expected):
        assert self.table[index].eval(self.state, self.model) == expected

    product_cases = [
        ((value, value, value, value), pytest.approx(1.61061273e-07))
        for value in [set_const, set_expr, set_var]
    ]

    for value in [set_const, set_expr, set_var]:
        for zero in [zero_expr, zero_var, zero_resource_var, 0]:
            product_cases.append(((zero, zero, zero, value), pytest.approx(0.12)))

    @pytest.mark.parametrize("index, expected", product_cases)
    def test_product(self, index, expected):
        assert self.table.product(index).eval(self.state, self.model) == expected

    max_cases = [
        ((value, value, value, value), pytest.approx(0.4))
        for value in [set_const, set_expr, set_var]
    ]

    for value in [set_const, set_expr, set_var]:
        for zero in [zero_expr, zero_var, zero_resource_var, 0]:
            max_cases.append(((zero, zero, zero, value), pytest.approx(0.4)))

    @pytest.mark.parametrize("index, expected", max_cases)
    def test_max(self, index, expected):
        assert self.table.max(index).eval(self.state, self.model) == expected

    min_cases = [
        ((value, value, value, value), pytest.approx(0.2))
        for value in [set_const, set_expr, set_var]
    ]

    for value in [set_const, set_expr, set_var]:
        for zero in [zero_expr, zero_var, zero_resource_var, 0]:
            min_cases.append(((zero, zero, zero, value), pytest.approx(0.3)))

    @pytest.mark.parametrize("index, expected", min_cases)
    def test_min(self, index, expected):
        assert self.table.min(index).eval(self.state, self.model) == expected

    error_cases = [
        (-1, 0, 0),
        (0, -1, 0),
        (0, 0, -1),
        (dp.FloatExpr(1), 0, 0),
        (0, dp.FloatExpr(1), 0),
        (0, 0, dp.FloatExpr(1)),
    ]

    @pytest.mark.parametrize("index", error_cases)
    def test_error(self, index):
        with pytest.raises(BaseException):
            self.table[index].eval(self.state, self.model)



def read(filename):
    with open(filename) as f:
        line = f.readline().strip()
        while line != "<end>":
            if line == "<number of tasks>":
                number_of_tasks = int(f.readline().strip())
            if line == "<cycle time>":
                cycle_time = int(f.readline().strip())
            if line == "<task times>":
                task_times = {}
                for _ in range(number_of_tasks):
                    pair = f.readline().strip().split()
                    task_times[int(pair[0])] = int(pair[1])
            if line == "<precedence relations>":
                predecessors = {i: [] for i in range(1, number_of_tasks + 1)}
                followers = {i: [] for i in range(1, number_of_tasks + 1)}
                relation = f.readline().strip()
                while len(relation) > 0:
                    relation = relation.split(",")
                    predecessor = int(relation[0])
                    follower = int(relation[1])
                    predecessors[follower].append(predecessor)
                    followers[predecessor].append(follower)
                    relation = f.readline().strip()

            line = f.readline().strip()

    return number_of_tasks, cycle_time, task_times, predecessors, followers


def validate(number_of_tasks, cycle_time, task_times, predecessors, solution, cost):
    if cost != len(solution):
        print(
            "The cost of solution {} mismatches the actual cost {}".format(
                cost, len(solution)
            )
        )
    scheduled = {}
    for (i, tasks) in enumerate(solution):
        time = 0
        for j in tasks:
            if j < 1 or j > number_of_tasks:
                print("task {} in station {} does not exist".format(j, i))
                return False
            if j in scheduled:
                print(
                    "task {} in station {} is already scheduled in station {}".format(
                        j, i, scheduled[j]
                    )
                )
                return False
            for k in predecessors[j]:
                if k not in scheduled and k not in tasks:
                    print(
                        "task {}, which is a predecessor of task {} in station {}, is not scheduled".format(
                            k, j, i
                        )
                    )
                    return False
            scheduled[j] = i
            time += task_times[j]
        if time > cycle_time:
            print(
                "station {} has total time {}, which exceeds the cycle time".format(
                    i + 1, time
                )
            )
            return False

    if len(scheduled) != number_of_tasks:
        print(
            "The number of scheduled tasks is {}, but should be {}".format(
                len(scheduled), number_of_tasks
            )
        )

    return True



def read(filename):
    with open(filename) as f:
        values = f.read().split()

    position = 0
    n = int(values[position])
    position += 1
    nodes = list(range(n))
    edges = {}
    for i in range(n):
        for j in range(n):
            if i != j:
                try:
                    edges[i, j] = int(values[position])
                except ValueError:
                    edges[i, j] = float(values[position])
            position += 1
    a = {}
    b = {}
    for i in range(n):
        try:
            a[i] = int(values[position])
        except ValueError:
            a[i] = float(values[position])
        position += 1
        try:
            b[i] = int(values[position])
        except ValueError:
            b[i] = float(values[position])
        position += 1

    return n, nodes, edges, a, b


def reduce_time_window(nodes, edges, a, b):
    reduced_a = dict(a)
    reduced_b = dict(b)
    change = True
    while change:
        change = False
        for i in nodes:
            a_i = min(
                reduced_b[i],
                min((reduced_a[j] + edges[j, i]) for j in nodes if (j, i) in edges),
                min((reduced_a[j] - edges[i, j]) for j in nodes if (i, j) in edges),
            )
            if a_i > reduced_a[i]:
                if not change:
                    change = True
                reduced_a[i] = a_i
            b_i = max(
                reduced_a[i],
                max((reduced_b[j] + edges[j, i]) for j in nodes if (j, i) in edges),
                max((reduced_b[j] - edges[i, j]) for j in nodes if (i, j) in edges),
            )
            if b_i < reduced_b[i]:
                if not change:
                    change = True
                reduced_b[i] = b_i

    return reduced_a, reduced_b


def reduce_edges(nodes, edges, a, b):
    print("edges: {}".format(len(edges)))
    forward_dependent = []
    for (i, j) in edges.keys():
        if i == 0 or j == 0 or a[i] <= b[j]:
            forward_dependent.append((i, j))

    direct_forward_dependent = {}
    for (i, j) in forward_dependent:
        if (
            i == 0
            or j == 0
            or all(b[i] > a[k] or b[k] > a[j] for k in nodes if k != 0)
            or (a[i] == a[j] and b[i] == b[j])
        ):
            direct_forward_dependent[i, j] = edges[i, j]

    print("reduced edges: {}".format(len(direct_forward_dependent)))
    return direct_forward_dependent


def validate(n, edges, a, b, solution, cost, tolerance=1e-4):
    previous = solution[0]
    if previous != 0:
        print(
            "The tour does not start from the depot {} but from {}".format(0, previous)
        )
        return False

    time = 0
    actual_cost = 0
    visited = set([0])

    for i in solution[1:-1]:
        if i < 0 or i > n - 1:
            print("Customer {} does not exist".format(i))
            return False
        if i in visited:
            print("Customer {} is already visited".format(i))
            return False
        visited.add(i)

        actual_cost += edges[previous, i]
        time = max(a[i], time + edges[previous, i])
        if time > b[i]:
            print("The time {} exceeds the deadline {} for {}".format(time, b[i], i))

        previous = i

    if solution[-1] != 0:
        print(
            "The tour does not return to the depot {}, but to {}".format(
                0, solution[-1]
            )
        )
        return False

    actual_cost += edges[previous, 0]

    if len(visited) != n:
        print(
            "The number of visited customers is {}, but should be {}".format(
                len(visited), n
            )
        )
        return False

    if abs(actual_cost - cost) > tolerance:
        print(
            "The cost of the solution {} mismatches the actual cost {}".format(
                cost, actual_cost
            )
        )
        return False

    return True
